# Configuration file for the Sphinx documentation builder.
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

# Find the path to the source files we want to to document, relative to the location of this file,
# convert it to an absolute path.
py_path = os.path.join(os.getcwd(), os.path.dirname(__file__), "../src")
sys.path.insert(0, os.path.abspath(py_path))

# -- Project information -----------------------------------------------------

project = "IQM client"
copyright = "2021-2025, IQM"
author = "IQM client developers"

# The short X.Y version.
version = ""
# The full version, including alpha/beta/rc tags.
release = ""
try:
    from iqm.iqm_client import __version__ as version
except ImportError:
    pass
else:
    release = version

# -- General configuration ---------------------------------------------------

# require a recent version of Sphinx
needs_sphinx = "7.2"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.todo",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# Include extra files in the HTML docs.
html_extra_path = []

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", ".*"]

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = "%Y-%m-%d"

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = True


# -- Autodoc ------------------------------------------------------------

# member ordering in autodoc output (default: 'alphabetical')
autodoc_member_order = "bysource"

# where should signature annotations appear in the docs, function signature or parameter description?
autodoc_typehints = "both"
# autodoc_typehints = 'description' puts the __init__ annotations into its docstring,
# which we thus have to include in the class documentation.
autoclass_content = "class"

# Sphinx 3.3+: manually clean up type alias rendering in the docs
# autodoc_type_aliases = {'TypeAlias': 'exa.experiment.somemodule.TypeAlias'}


# -- Autosummary ------------------------------------------------------------

# use autosummary to generate stub pages for API docs
autosummary_generate = True


# -- Options for HTML output -------------------------------------------------

import sphinx_book_theme

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_book_theme"

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [sphinx_book_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {}

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/images/logo.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/images/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%Y-%m-%d"

# Output file base name for HTML help builder.
htmlhelp_basename = "iqm_client-doc"


# -- MathJax options ----------------------------------------------------------

# Here we configure MathJax, mostly to define LaTeX macros.
mathjax3_config = {
    "tex": {
        "macros": {
            "vr": r"\vec{r}",  # no arguments
            "ket": [r"\left| #1 \right\rangle", 1],  # one argument
            "iprod": [r"\left\langle #1 | #2 \right\rangle", 2],  # two arguments
        }
    }
}


# -- External mapping ------------------------------------------------------------

python_version = ".".join(map(str, sys.version_info[0:2]))
intersphinx_mapping = {
    "python": ("https://docs.python.org/" + python_version, None),
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "qiskit": ("https://docs.quantum.ibm.com/api/qiskit", None),
}

use_local_target = os.getenv("USE_LOCAL_TARGET", "").lower()
if use_local_target != "true":
    extlinks = {
        "issue": ("https://jira.iqm.fi/browse/%s", "issue %s"),
        "mr": ("https://gitlab.iqm.fi/iqm/qccsw/exa/exa-repo/-/merge_requests/%s", "MR %s"),
    }


# -- Options for sphinxcontrib.bibtex -------------------------------------------------

# List of all bibliography files used.
# bibtex_bibfiles = ['references.bib']



# Copyright 2025 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Mocks server calls for testing
"""

from importlib.metadata import version
import json

from packaging.version import parse
import pytest
from requests import HTTPError, Response


@pytest.fixture()
def base_url() -> str:
    # NOTE: You should mock all HTTP requests in the tests, so we do not send out actual HTTP requests here!
    return "https://example.com"


class MockJsonResponse:
    def __init__(self, status_code: int, json_data: dict, history: list[Response] | None = None):
        self.status_code = status_code
        self.json_data = json_data
        self.history = history
        self.url = "https://example.com"

    @property
    def text(self):
        # NOTE cannot handle UUIDs
        return json.dumps(self.json_data)

    def json(self):
        return self.json_data

    def raise_for_status(self):
        if 400 <= self.status_code < 600:
            raise HTTPError(f"{self.status_code}", response=self)


def mock_supported_client_libraries_response(
    iqm_client_name: str = "iqm-client", max_version: str | None = None, min_version: str | None = None
) -> MockJsonResponse:
    client_version = parse(version("iqm-client"))
    min_version = f"{client_version.major}.0" if min_version is None else min_version
    max_version = f"{client_version.major + 1}.0" if max_version is None else max_version
    return MockJsonResponse(
        200,
        {
            iqm_client_name: {
                "name": iqm_client_name,
                "min": min_version,
                "max": max_version,
            }
        },
    )



# Copyright 2020–2022 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""DeviceMetadata subtype for IQM devices."""

from __future__ import annotations

from collections.abc import Iterable
from typing import FrozenSet, Optional

import cirq
from cirq import Gate, NamedQid, devices, ops
from cirq.contrib.routing.router import nx
from iqm.cirq_iqm.serialize import _IQM_CIRQ_OP_MAP
from iqm.iqm_client import DynamicQuantumArchitecture


@cirq.value.value_equality
class IQMDeviceMetadata(devices.DeviceMetadata):
    """Hardware metadata for IQM devices.

    Args:
        qubits: qubits on the device
        connectivity: qubit connectivity graph of the device
        operations: Supported quantum operations of the device, mapping op types to their possible loci.
        gateset: Native gateset of the device. If None, a default IQM device gateset will be used.
        resonators: computational resonators of the device
        architecture: architecture from which values of the other arguments were obtained

    """

    QUBIT_NAME_PREFIX: str = "QB"
    """prefix for qubit names, to be followed by their numerical index"""

    RESONATOR_DIMENSION: int = 2
    """Dimension abstraction for the resonator Qids"""

    def __init__(
        self,
        qubits: Iterable[NamedQid],
        connectivity: Iterable[tuple[NamedQid, ...]],
        *,
        operations: Optional[dict[type[cirq.Gate], list[tuple[cirq.NamedQid, ...]]]] = None,
        gateset: Optional[cirq.Gateset] = None,
        resonators: Iterable[NamedQid] = (),
        architecture: Optional[DynamicQuantumArchitecture] = None,
    ):
        """Construct an IQMDeviceMetadata object."""
        nx_graph = nx.Graph()
        for edge in connectivity:
            if len(edge) != 2:
                raise ValueError("Connectivity must be an iterable of 2-tuples.")
            nx_graph.add_edge(edge[0], edge[1])
        super().__init__(qubits, nx_graph)
        self._qubit_set: FrozenSet[NamedQid] = frozenset(qubits)
        self._resonator_set: FrozenSet[NamedQid] = frozenset(resonators)

        if gateset is None:
            if operations is None:
                # default gateset for IQM devices
                gateset = cirq.Gateset(
                    ops.PhasedXPowGate, ops.XPowGate, ops.YPowGate, ops.MeasurementGate, ops.CZPowGate
                )
                sqg_list: list[type[Gate]] = [ops.PhasedXPowGate, ops.XPowGate, ops.YPowGate, ops.MeasurementGate]
                operations = {}
                operations[ops.CZPowGate] = list(tuple(edge) for edge in connectivity)
                operations.update({gate: [(qb,) for qb in qubits] for gate in sqg_list})
            else:
                gateset = cirq.Gateset(*operations.keys())
        self._gateset = gateset

        if operations is None:
            raise ValueError("Operations must be provided if a gateset is provided, it cannot be reconstructed.")
        self.operations = operations

        self.architecture = architecture

    @property
    def resonator_set(self) -> FrozenSet[NamedQid]:
        """Returns the set of resonators on the device.

        Returns:
            Frozenset of resonators on device.

        """
        return self._resonator_set

    @classmethod
    def from_architecture(cls, architecture: DynamicQuantumArchitecture) -> IQMDeviceMetadata:
        """Returns device metadata object created based on dynamic quantum architecture"""
        qubits = tuple(NamedQid(qb, dimension=2) for qb in architecture.qubits)
        resonators = tuple(
            NamedQid(cr, dimension=cls.RESONATOR_DIMENSION) for cr in architecture.computational_resonators
        )

        def get_qid_locus(locus: Iterable[str]) -> tuple[NamedQid, ...]:
            """Converts locus component names to Qids."""
            return tuple(
                (
                    NamedQid(component, dimension=2)
                    if component in qubits
                    else NamedQid(component, dimension=cls.RESONATOR_DIMENSION)
                )
                for component in locus
            )

        # connectivity consists of all arity-2 gate loci in the DQA
        connectivity = tuple(
            get_qid_locus(locus)
            for gate_info in architecture.gates.values()
            for locus in gate_info.loci
            if len(locus) == 2
        )
        operations: dict[type[cirq.Gate], list[tuple[NamedQid, ...]]] = {
            cirq_op: [get_qid_locus(locus) for locus in gate_info.loci]
            for gate_name, gate_info in architecture.gates.items()
            if gate_name in _IQM_CIRQ_OP_MAP
            for cirq_op in _IQM_CIRQ_OP_MAP[gate_name]
        }
        return cls(
            qubits,
            connectivity,
            operations=operations,
            resonators=resonators,
            architecture=architecture,
        )

    @classmethod
    def from_qubit_indices(
        cls,
        qubit_count: int,
        connectivity_indices: Iterable[set[int]],
        gateset: Optional[tuple[type[cirq.Gate]]] = None,
    ) -> IQMDeviceMetadata:
        """Returns device metadata object created based on connectivity specified using qubit indices only."""
        qubits = tuple(NamedQid.range(1, qubit_count + 1, prefix=cls.QUBIT_NAME_PREFIX, dimension=2))
        if set(map(len, connectivity_indices)) != {2}:
            raise ValueError("connectivity_indices must be an iterable of 2-sets.")
        connectivity = tuple(
            tuple(NamedQid(f"{cls.QUBIT_NAME_PREFIX}{qb}", dimension=2) for qb in edge) for edge in connectivity_indices
        )
        if gateset:
            sqg_list: list[type[Gate]] = [
                g for g in gateset if g in [ops.PhasedXPowGate, ops.XPowGate, ops.YPowGate, ops.MeasurementGate]
            ]
            operations: dict[type[cirq.Gate], list[tuple[cirq.NamedQid, ...]]] = {}
            if ops.CZPowGate in gateset:
                operations[ops.CZPowGate] = list(tuple(edge) for edge in connectivity)
            if ops.ISwapPowGate in gateset:
                operations[ops.ISwapPowGate] = list(tuple(edge) for edge in connectivity)
            operations.update({gate: [(qb,) for qb in qubits] for gate in sqg_list})
            return cls(qubits, connectivity, operations=operations, gateset=cirq.Gateset(*gateset))
        return cls(qubits, connectivity)

    @property
    def gateset(self) -> cirq.Gateset:
        """Returns the ``cirq.Gateset`` of supported gates on this device."""
        return self._gateset

    def _value_equality_values_(self):
        return *super()._value_equality_values_(), self._gateset



# Copyright 2020–2022 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Demonstrates importing a quantum circuit from an OpenQASM 2.0 file, transforming it into
the Apollo native gateset and connectivity, and then executing it on a simulator.
"""

import cirq
from demo_common import demo
from iqm.cirq_iqm.devices import Apollo
from iqm.cirq_iqm.extended_qasm_parser import circuit_from_qasm


def demo_apollo(use_qsim: bool = False) -> None:
    """Run the demo using the Apollo architecture."""
    print("\nApollo demo\n===========\n")
    device = Apollo()
    qasm_program = """
        OPENQASM 2.0;
        include "qelib1.inc";

        qreg q[6];
        creg meas[6];

        U(0.2, 0.5, 1.7) q[1];
        h q[0];
        h q[2];
        h q[3];
        h q[4];
        h q[5];
        cx q[1], q[2];
        cx q[2], q[5];
        cx q[3], q[4];
        cx q[0], q[1];
        measure q -> meas;
    """
    circuit = circuit_from_qasm(qasm_program)

    # add some more gates
    q2 = cirq.NamedQubit("q_2")
    q3 = cirq.NamedQubit("q_3")
    circuit.insert(len(circuit) - 2, cirq.CXPowGate(exponent=0.723)(q2, q3))

    demo(device, circuit, use_qsim=use_qsim)


if __name__ == "__main__":
    demo_apollo()



# Copyright 2024 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains definitions of IQM Server API endpoints."""

from enum import Enum, auto
from posixpath import join
import warnings


class APIEndpoint(Enum):
    """Supported API endpoints."""

    CONFIGURATION = auto()
    SUBMIT_JOB = auto()
    GET_JOB_REQUEST_PARAMETERS = auto()
    GET_JOB_RESULT = auto()
    GET_JOB_STATUS = auto()
    GET_JOB_CALIBRATION_SET_ID = auto()
    GET_JOB_CIRCUITS_BATCH = auto()
    GET_JOB_TIMELINE = auto()
    GET_JOB_ERROR_LOG = auto()
    GET_JOB_COUNTS = auto()
    ABORT_JOB = auto()
    DELETE_JOB = auto()
    HEALTH = auto()
    ABOUT = auto()
    CLIENT_LIBRARIES = auto()

    # Calibration and Calibration Service endpoints
    CALIBRATION_SERVICE_CONFIGURATION = auto()
    QUANTUM_ARCHITECTURE = auto()
    STATIC_QUANTUM_ARCHITECTURE = auto()
    CHANNEL_PROPERTIES = auto()
    QUALITY_METRICS_LATEST = auto()
    QUALITY_METRICS = auto()
    QUALITY_METRICS_MONITORING = auto()
    CALIBRATED_GATES = auto()
    START_CALIBRATION_JOB = auto()
    ABORT_CALIBRATION_JOB = auto()
    CALIBRATION_SERVICE_JOBS = auto()
    CALIBRATION = auto()


class APIVariant(Enum):
    """Supported API versions and variants."""

    V1 = "V1"  # QCCSW Cocos-based circuits execution
    V2 = "V2"  # Station-Control-based circuits execution
    RESONANCE_COCOS_V1 = "RESONANCE_COCOS_V1"  # IQM Resonance Cocos API


class APIConfig:
    """Provides supported API endpoints for a given API variant."""

    def __init__(self, variant: APIVariant, iqm_server_url: str):
        """Args:
        variant: API variant.
        iqm_server_url: URL of the IQM server,
                        e.g. https://test.qc.iqm.fi/cocos or https://cocos.resonance.meetiqm.com/garnet for .V1
                        or https://test.qc.iqm.fi for .V2

        """
        self.variant = variant

        # V1 API is deprecated and will be removed in a future release.
        # V2 API is supported by the QCCSW starting with version 3.3.0.
        if self.variant == APIVariant.V1:
            warnings.warn(
                "The V1 API is deprecated and will be removed in a future release. Please use the V2 API instead.",
                DeprecationWarning,
            )

        self.iqm_server_url = iqm_server_url
        self.urls = self._get_api_urls()

    def _get_api_urls(self) -> dict[APIEndpoint, str]:
        """Returns:
        Relative URLs for each supported API endpoints.

        """
        if self.variant == APIVariant.RESONANCE_COCOS_V1:
            return {
                APIEndpoint.SUBMIT_JOB: "jobs",
                APIEndpoint.GET_JOB_RESULT: "jobs/%s",
                APIEndpoint.GET_JOB_STATUS: "jobs/%s/status",
                APIEndpoint.GET_JOB_COUNTS: "jobs/%s/counts",
                APIEndpoint.ABORT_JOB: "jobs/%s/abort",
                APIEndpoint.QUANTUM_ARCHITECTURE: "quantum-architecture",
                APIEndpoint.CALIBRATED_GATES: "api/v1/calibration/%s/gates",
                APIEndpoint.CLIENT_LIBRARIES: "info/client-libraries",
            }
        if self.variant == APIVariant.V1:
            return {
                APIEndpoint.CONFIGURATION: "configuration",
                APIEndpoint.QUALITY_METRICS_LATEST: "calibration/metrics/latest",
                APIEndpoint.QUALITY_METRICS: "calibration/metrics/%s",
                APIEndpoint.SUBMIT_JOB: "jobs",
                APIEndpoint.GET_JOB_RESULT: "jobs/%s",
                APIEndpoint.GET_JOB_STATUS: "jobs/%s/status",
                APIEndpoint.GET_JOB_COUNTS: "jobs/%s/counts",
                APIEndpoint.ABORT_JOB: "jobs/%s/abort",
                APIEndpoint.ABORT_CALIBRATION_JOB: "jobs/%s/abort",
                APIEndpoint.DELETE_JOB: "jobs/%s",
                APIEndpoint.QUANTUM_ARCHITECTURE: "quantum-architecture",
                APIEndpoint.STATIC_QUANTUM_ARCHITECTURE: "api/v1/quantum-architecture",
                APIEndpoint.CALIBRATED_GATES: "api/v1/calibration/%s/gates",
                APIEndpoint.QUALITY_METRICS_MONITORING: "api/v1/monitor/calibration/metrics",
                APIEndpoint.HEALTH: "health",
                APIEndpoint.ABOUT: "about",
                APIEndpoint.CLIENT_LIBRARIES: "info/client-libraries",
                APIEndpoint.START_CALIBRATION_JOB: "calibration/run",
                APIEndpoint.CALIBRATION_SERVICE_CONFIGURATION: "calibration/configuration",
                APIEndpoint.CALIBRATION_SERVICE_JOBS: "calibration/jobs",
                APIEndpoint.CALIBRATION: "api/v1/calibration/%s",
            }
        if self.variant == APIVariant.V2:
            return {
                APIEndpoint.GET_JOB_REQUEST_PARAMETERS: "station/circuits/%s/request_parameters",
                APIEndpoint.CONFIGURATION: "cocos/configuration",
                APIEndpoint.QUALITY_METRICS_LATEST: "cocos/calibration/metrics/latest",
                APIEndpoint.QUALITY_METRICS: "cocos/calibration/metrics/%s",
                APIEndpoint.SUBMIT_JOB: "station/circuits",
                APIEndpoint.GET_JOB_RESULT: "station/circuits/%s/measurements",
                APIEndpoint.GET_JOB_STATUS: "station/circuits/%s/status",
                APIEndpoint.GET_JOB_CALIBRATION_SET_ID: "station/circuits/%s/calibration_set_id",
                APIEndpoint.GET_JOB_CIRCUITS_BATCH: "station/circuits/%s/circuits_batch",
                APIEndpoint.GET_JOB_TIMELINE: "station/circuits/%s/timeline",
                APIEndpoint.GET_JOB_ERROR_LOG: "station/circuits/%s/error_log",
                APIEndpoint.GET_JOB_COUNTS: "station/circuits/%s/counts",
                APIEndpoint.ABORT_JOB: "station/circuits/%s/abort",
                APIEndpoint.ABORT_CALIBRATION_JOB: "cocos/jobs/%s/abort",
                APIEndpoint.DELETE_JOB: "station/circuits/%s",
                APIEndpoint.QUANTUM_ARCHITECTURE: "cocos/quantum-architecture",
                APIEndpoint.STATIC_QUANTUM_ARCHITECTURE: "cocos/api/v1/quantum-architecture",
                APIEndpoint.CHANNEL_PROPERTIES: "station/channel-properties",
                APIEndpoint.CALIBRATED_GATES: "cocos/api/v1/calibration/%s/gates",
                APIEndpoint.QUALITY_METRICS_MONITORING: "cocos/api/v1/monitor/calibration/metrics",
                APIEndpoint.HEALTH: "cocos/health",
                APIEndpoint.ABOUT: "cocos/about",
                APIEndpoint.CLIENT_LIBRARIES: "info/client-libraries",
                APIEndpoint.START_CALIBRATION_JOB: "cocos/calibration/run",
                APIEndpoint.CALIBRATION_SERVICE_CONFIGURATION: "cocos/calibration/configuration",
                APIEndpoint.CALIBRATION_SERVICE_JOBS: "cocos/calibration/jobs",
                APIEndpoint.CALIBRATION: "cocos/api/v1/calibration/%s",
            }
        raise ValueError(f"Unsupported API variant: {self.variant}")

    def is_supported(self, endpoint: APIEndpoint) -> bool:
        """Args:
            endpoint: API endpoint.

        Returns:
            True if the endpoint is supported, False otherwise.

        """
        return endpoint in self.urls

    def url(self, endpoint: APIEndpoint, *args) -> str:
        """Args:
            endpoint: API endpoint.
            args: Arguments to be passed to the URL.

        Returns:
            URL for the given endpoint.

        Raises:
            ValueError: If the endpoint is not supported.

        """
        url = self.urls.get(endpoint)
        if url is None:
            raise ValueError(f"Unsupported API endpoint: {endpoint}")
        return join(self.iqm_server_url, url % args)



# Copyright 2020–2022 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Describes IQM quantum architectures in the Cirq framework.

The description includes the qubit connectivity, the native gate set, and the gate decompositions
to use with the architecture.
"""

from __future__ import annotations

import collections.abc as ca
from math import pi as PI
from typing import Optional, Sequence, cast

import cirq
from cirq import devices, ops, protocols
from cirq.contrib.routing.router import nx
from iqm.cirq_iqm.iqm_gates import IQMMoveGate
from iqm.cirq_iqm.transpiler import transpile_insert_moves_into_circuit

from .iqm_device_metadata import IQMDeviceMetadata


def _verify_unique_measurement_keys(operations: ca.Iterable[cirq.Operation]) -> None:
    """Raises an error if a measurement key is repeated in the given set of Operations."""
    seen_keys: set[str] = set()
    for op in operations:
        if protocols.is_measurement(op):
            key = protocols.measurement_key_name(op)
            if key in seen_keys:
                raise ValueError(f"Measurement key {key} repeated")
            seen_keys.add(key)


class IQMDevice(devices.Device):
    """ABC for the properties of a specific IQM quantum architecture.

    Adds extra functionality on top of the basic :class:`cirq.Device` class for decomposing gates,
    optimizing circuits and mapping qubits.

    Args:
        metadata: device metadata which contains the qubits, their connectivity, and the native gateset

    """

    def __init__(self, metadata: IQMDeviceMetadata):
        self._metadata = metadata
        self.qubits = tuple(sorted(self._metadata.qubit_set))
        self.resonators = tuple(sorted(self._metadata.resonator_set))
        self.supported_operations = self._metadata.operations

    @property
    def metadata(self) -> IQMDeviceMetadata:
        """Returns metadata for the device."""
        return self._metadata

    @classmethod
    def get_qubit_index(cls, qubit: cirq.NamedQubit) -> int:
        """The numeric index of the given qubit on the device."""
        return int(qubit.name[len(IQMDeviceMetadata.QUBIT_NAME_PREFIX) :])

    def get_qubit(self, index: int) -> cirq.Qid:
        """The device qubit corresponding to the given numeric index."""
        return self.qubits[index - 1]  # 1-based indexing

    def check_qubit_connectivity(self, operation: cirq.Operation) -> None:
        """Raises a ValueError if operation acts on qubits that are not connected."""
        if len(operation.qubits) >= 2 and not self.has_valid_operation_targets(operation):
            raise ValueError(f"Unsupported qubit connectivity required for {operation!r}")

    def is_native_operation(self, op: cirq.Operation) -> bool:
        """Predicate, True iff the given operation is considered native for the architecture."""
        check = (
            isinstance(op, (ops.GateOperation, ops.TaggedOperation))
            and (op.gate is not None)
            and (op.gate in self._metadata.gateset)
        )
        if check and isinstance(op.gate, ops.CZPowGate):
            return op.gate.exponent == 1
        return check

    def has_valid_operation_targets(self, op: cirq.Operation) -> bool:
        """Predicate, True iff the given operation is native and its targets are valid."""
        matched_support = [
            (g, loci)
            for g, loci in self.supported_operations.items()
            if op.gate is not None and op.gate in cirq.GateFamily(g)
        ]
        if len(matched_support) > 0:
            gf, valid_loci = matched_support[0]
            if gf == cirq.MeasurementGate:  # Measurements can be done on any available qubits
                valid_qubits = set(q for locus in valid_loci for q in locus)
                return all(q in valid_qubits for q in op.qubits)
            if issubclass(gf, cirq.InterchangeableQubitsGate):
                return any(set(t) == set(op.qubits) for t in valid_loci)
            return op.qubits in valid_loci
        return False

    def operation_decomposer(self, op: cirq.Operation) -> Optional[list[cirq.Operation]]:
        """Decomposes operations into the native operation set.

        All the decompositions below keep track of global phase (required for decomposing controlled gates).

        Args:
            op: operation to decompose

        Returns:
            decomposition, or None to pass ``op`` to the Cirq native decomposition machinery instead

        """
        # It seems that Cirq native decompositions ignore global phase entirely?

        PI_2 = PI / 2

        # common gates used in gate decompositions
        CZ = ops.CZPowGate()
        Lx = ops.rx(PI_2)
        Lxi = ops.rx(-PI_2)
        Ly = ops.ry(PI_2)
        Lyi = ops.ry(-PI_2)

        if isinstance(op.gate, ops.CZPowGate):
            # decompose CZPowGate using ZZPowGate
            t = op.gate.exponent
            s = op.gate.global_shift
            L = ops.rz(t / 2 * PI)
            return [
                ops.ZZPowGate(exponent=-0.5 * t, global_shift=-2 * s - 1).on(*op.qubits),
                L.on(op.qubits[0]),
                L.on(op.qubits[1]),
            ]
        if isinstance(op.gate, ops.ZZPowGate):
            # decompose ZZPowGate using two CZs
            t = op.gate.exponent
            s = op.gate.global_shift
            return [
                Lyi.on(op.qubits[1]),
                CZ.on(*op.qubits),
                ops.XPowGate(exponent=-t, global_shift=-1 - s).on(op.qubits[1]),
                CZ.on(*op.qubits),
                Ly.on(op.qubits[1]),
            ]
        if isinstance(op.gate, ops.ISwapPowGate):
            # decompose ISwapPowGate using two CZs
            t = op.gate.exponent
            s = op.gate.global_shift
            x = -0.5 * t
            return [
                Lxi.on(op.qubits[0]),
                Lxi.on(op.qubits[1]),
                Lyi.on(op.qubits[1]),
                CZ.on(*op.qubits),
                ops.XPowGate(exponent=x, global_shift=-0.5 - 2 * s).on(op.qubits[0]),
                ops.XPowGate(exponent=-x, global_shift=-0.5).on(op.qubits[1]),
                CZ.on(*op.qubits),
                Ly.on(op.qubits[1]),
                Lx.on(op.qubits[0]),
                Lx.on(op.qubits[1]),
            ]
        if isinstance(op.gate, ops.ZPowGate):
            # Rz using Rx, Ry
            q = op.qubits[0]
            return [
                ops.XPowGate(exponent=-0.5).on(q),
                ops.YPowGate(exponent=op.gate.exponent).on(q),
                ops.XPowGate(exponent=0.5).on(q),
            ]
        return None

    def decompose_operation(self, operation: cirq.Operation) -> cirq.OP_TREE:
        """Decompose a single quantum operation into the native operation set."""
        if self.is_native_operation(operation):
            return operation

        return protocols.decompose(
            operation,
            intercepting_decomposer=self.operation_decomposer,
            keep=self.is_native_operation,
            on_stuck_raise=None,
        )

    def route_circuit(
        self,
        circuit: cirq.Circuit,
        *,
        initial_mapper: Optional[cirq.AbstractInitialMapper] = None,
        qubit_subset: Optional[Sequence[cirq.Qid]] = None,
    ) -> tuple[cirq.Circuit, dict[cirq.Qid, cirq.Qid], dict[cirq.Qid, cirq.Qid]]:
        """Routes the given circuit to the device connectivity and qubit names.

        The routed circuit uses the device qubits, and may have additional SWAP gates inserted to perform the qubit
        routing. The transformer :class:`cirq.RouterCQC` is used for routing.
        Note that only gates of one or two qubits, and measurement operations of arbitrary size are supported.

        Adds the attribute ``iqm_calibration_set_id`` to the routed circuit, with value taken from
        ``self.metadata.architecture.calibration_set_id`` if available, otherwise None.

        Args:
            circuit: Circuit to route.
            initial_mapper: Initial mapping from ``circuit`` qubits to device qubits, to serve as
                the starting point of the routing. ``None`` means it will be generated automatically.
            qubit_subset: Restrict the routing to this subset of the device qubits. If ``None``,
                use the entire device.

        Returns:
            routed circuit, initial mapping before inserting SWAP gates (see :func:`cirq.RouterCQC.route_circuit`),
            final mapping from physical qubits to physical qubits (see :func:`cirq.RouterCQC.route_circuit`)

        Raises:
            ValueError: routing is impossible

        """
        modified_circuit = circuit.copy()

        if self.metadata.resonator_set:
            # If the device has computational resonators, we use a modified connection graph for routing
            # by adding edges between all qubits connected to the same resonator.
            move_routing = True
            graph = nx.Graph()
            for edge in self.metadata.nx_graph.edges:
                q, r = edge if edge[1] in self.resonators else edge[::-1]
                if r not in self.resonators:
                    graph.add_edge(*edge)
                else:
                    for n in self.metadata.nx_graph.neighbors(r):
                        if n != q and not graph.has_edge(q, n) and not graph.has_edge(n, q):
                            graph.add_edge(q, n)
        else:
            graph = self._metadata.nx_graph
            move_routing = False

        if qubit_subset is not None:
            graph = graph.subgraph(qubit_subset)

        # Route the modified circuit.
        router = cirq.RouteCQC(graph)
        routed_circuit, initial_mapping, final_mapping = router.route_circuit(
            modified_circuit, initial_mapper=initial_mapper
        )
        routed_circuit = cast(cirq.Circuit, routed_circuit)
        # TODO routing can apply SWAP gates even after formerly terminal single-qubit measurements.
        # It would make sense to commute single-qubit measurements through SWAPs as far towards the
        # end of the circuit as possible.

        if move_routing:
            # Decompose the SWAP  gates to the native gate set.
            routed_circuit = self.decompose_circuit(routed_circuit)
            # Insert IQMMoveGates into the circuit.
            routed_circuit = transpile_insert_moves_into_circuit(routed_circuit, self)

        routed_circuit.iqm_calibration_set_id = (  # type: ignore
            self._metadata.architecture.calibration_set_id if self._metadata.architecture is not None else None
        )

        return routed_circuit, initial_mapping, final_mapping

    def decompose_circuit(self, circuit: cirq.Circuit) -> cirq.Circuit:
        """Decomposes the given circuit to the native gate set of the device.

        Adds the attribute ``iqm_calibration_set_id`` to the decomposed circuit, with value taken from
        ``self.metadata.architecture.calibration_set_id`` if available, otherwise None.

        Args:
            circuit: circuit to decompose

        Returns:
            decomposed circuit

        """
        moments = ops.transform_op_tree(
            circuit.moments,
            self.decompose_operation,
            preserve_moments=False,
        )
        decomposed_circuit = cirq.Circuit(moments)
        decomposed_circuit.iqm_calibration_set_id = (  # type: ignore
            self._metadata.architecture.calibration_set_id if self._metadata.architecture is not None else None
        )
        return decomposed_circuit

    def validate_circuit(self, circuit: cirq.AbstractCircuit) -> None:
        super().validate_circuit(circuit)
        _verify_unique_measurement_keys(circuit.all_operations())
        self.validate_moves(circuit)

    def validate_operation(self, operation: cirq.Operation) -> None:
        if not isinstance(operation.untagged, cirq.GateOperation):
            raise ValueError(f"Unsupported operation: {operation!r}")

        if not self.is_native_operation(operation):
            raise ValueError(f"Unsupported gate type: {operation.gate!r}")

        for qubit in operation.qubits:
            if qubit not in self.qubits and qubit not in self.resonators:
                raise ValueError(f"Qubit not on device: {qubit!r}")

        if not self.has_valid_operation_targets(operation):
            raise ValueError(f"Unsupported operation between qubits: {operation!r}")

    def validate_move(self, operation: cirq.Operation) -> None:
        """Validates whether the IQMMoveGate is between qubit and resonator registers.

        Args:
            operation (cirq.Operation): Operation to check

        Raises:
            ValueError: In case the the first argument of the IQMMoveGate is not a qubit,
                        or if the second argument is not a resonator on this device.

        Returns:
            None when the IQMMoveGate is used correctly.

        """
        if isinstance(operation.gate, IQMMoveGate):
            if operation.qubits[0] not in self.qubits:
                raise ValueError(
                    f"IQMMoveGate is only supported with a qubit register as the first argument, \
                    but got {operation.qubits[0]!r}"
                )
            if operation.qubits[1] not in self.resonators:
                raise ValueError(
                    f"IQMMoveGate is only supported with a resonator register as the second argument, \
                    but got {operation.qubits[1]!r}"
                )

    def validate_moves(self, circuit: cirq.AbstractCircuit) -> None:
        """Validates whether the IQMMoveGates are correctly applied in the circuit.

        Args:
            circuit (cirq.AbstractCircuit): The circuit to validate.

        Raises:
            ValueError: If the IQMMoveGate is applied incorrectly.

        Returns:
            None if the IQMMoveGates are applied correctly.

        """
        moves: dict[cirq.Qid, list[cirq.Qid]] = {r: [] for r in self.resonators}
        for moment in circuit:
            for operation in moment.operations:
                if isinstance(operation.gate, IQMMoveGate):
                    self.validate_move(operation)
                    moves[operation.qubits[1]].append(operation.qubits[0])
        for res, qubits in moves.items():
            qubits_updated = qubits
            while len(qubits_updated) > 1:
                q1, q2, *rest = qubits_updated
                if q1 != q2:
                    raise ValueError(f"IQMMoveGate({q2!r}, {res!r}) is applied between two logical qubit states.")
                qubits_updated = rest
            if len(qubits_updated) != 0:
                raise ValueError(f"Circuit ends with a qubit state in the resonator {res!r}.")

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self._metadata == other._metadata



# Copyright 2020–2021 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cirq adapter for IQM's quantum computers."""

from importlib.metadata import PackageNotFoundError, version
import warnings

from .devices import *  # noqa: F403*
from .extended_qasm_parser import circuit_from_qasm

try:
    DIST_NAME = "cirq-iqm"
    __version__ = version(DIST_NAME)
except PackageNotFoundError:
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError
from .iqm_gates import *  # noqa: F403
from .transpiler import transpile_insert_moves_into_circuit

warnings.warn(
    DeprecationWarning(
        "The cirq-iqm package is deprecated and new versions of Cirq on IQM will be published as part of "
        "iqm-client. Please uninstall cirq-iqm and install iqm-client[cirq] to get the newest version."
    )
)



# Copyright 2023 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Transpilation tool to optimize the decomposition of single-qubit gates tailored to IQM hardware."""

import math
import warnings

import numpy as np
from qiskit import QuantumCircuit
from qiskit.circuit.equivalence_library import SessionEquivalenceLibrary
from qiskit.circuit.library import RGate, UnitaryGate
from qiskit.dagcircuit import DAGCircuit
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.passes import BasisTranslator, Optimize1qGatesDecomposition, RemoveBarriers
from qiskit.transpiler.passmanager import PassManager

TOLERANCE = 1e-10  # The tolerance for equivalence checking against zero.


class IQMOptimizeSingleQubitGates(TransformationPass):
    r"""Optimize the decomposition of single-qubit gates for the IQM gate set.

    This optimization pass expects the circuit to be correctly layouted and translated to the IQM architecture
    and raises an error otherwise.
    The optimization logic follows the steps:

    1. Convert single-qubit gates to :math:`U` gates and combine all neighboring :math:`U` gates.
    2. Convert :math:`U` gates according to
       :math:`U(\theta , \phi , \lambda) = ~ RZ(\phi + \lambda) R(\theta, \pi / 2  - \lambda)`.
    3. Commute `RZ` gates to the end of the circuit using the fact that `RZ` and `CZ` gates commute, and
       :math:`R(\theta , \phi) RZ(\lambda) = RZ(\lambda) R(\theta, \phi - \lambda)`.
    4. Drop `RZ` gates immediately before measurements, and otherwise replace them according to
       :math:`RZ(\lambda) = R(\pi, \lambda / 2) R(- \pi, 0)`.

    Args:
        drop_final_rz: Drop terminal RZ gates even if there are no measurements following them (since they do not affect
            the measurement results). Note that this will change the unitary propagator of the circuit.
            It is recommended always to set this to true as the final RZ gates do no change the measurement outcomes of
            the circuit.
        ignore_barriers (bool): Removes the barriers from the circuit before optimization (default = False).

    """

    def __init__(self, drop_final_rz: bool = True, ignore_barriers: bool = False):
        super().__init__()
        self._basis = ["r", "cz", "move"]
        self._intermediate_basis = ["u", "cz", "move"]
        self._drop_final_rz = drop_final_rz
        self._ignore_barriers = ignore_barriers

    def run(self, dag: DAGCircuit) -> DAGCircuit:
        self._validate_ops(dag)
        # accumulated RZ angles for each qubit, from the beginning of the circuit to the current gate
        rz_angles: list[float] = [0] * dag.num_qubits()

        if self._ignore_barriers:
            dag = RemoveBarriers().run(dag)
        # convert all gates in the circuit to U and CZ gates
        dag = BasisTranslator(SessionEquivalenceLibrary, self._intermediate_basis).run(dag)
        # combine all sequential U gates into one
        dag = Optimize1qGatesDecomposition(self._intermediate_basis).run(dag)
        for node in dag.topological_op_nodes():
            if node.name == "u":
                # convert into PRX + RZ
                qubit_index = dag.find_bit(node.qargs[0])[0]
                if isinstance(node.op.params[0], float) and math.isclose(node.op.params[0], 0, abs_tol=TOLERANCE):
                    dag.remove_op_node(node)
                else:
                    dag.substitute_node(
                        node, RGate(node.op.params[0], np.pi / 2 - node.op.params[2] - rz_angles[qubit_index])
                    )
                phase = node.op.params[1] + node.op.params[2]
                dag.global_phase += phase / 2
                rz_angles[qubit_index] += phase
            elif node.name in {"measure", "reset"}:
                # measure and reset destroy phase information. The local phases before and after such
                # an operation are in principle independent, and the local computational frame phases
                # are arbitrary so we could set rz_angles to any values here, but zeroing the
                # angles results in fewest changes to the circuit.
                for qubit in node.qargs:
                    rz_angles[dag.find_bit(qubit)[0]] = 0
            elif node.name == "barrier":
                # TODO barriers are meant to restrict circuit optimization, so strictly speaking
                # we should output any accumulated ``rz_angles`` here as explicit z rotations (like
                # the final rz:s). However, ``rz_angles`` simply represents a choice of phases for the
                # local computational frames for the rest of the circuit (the initial part has already
                # been transformed). This choice of local phases is in principle arbitrary, so maybe it
                # makes no sense to convert it into active z rotations if we hit a barrier?
                pass
            elif node.name == "move":
                # acts like iSWAP with RZ, moving it to the other component
                qb, res = dag.find_bit(node.qargs[0])[0], dag.find_bit(node.qargs[1])[0]
                rz_angles[res], rz_angles[qb] = rz_angles[qb], rz_angles[res]
            elif node.name in {"cz", "delay"}:
                pass  # commutes with RZ gates
            else:
                raise ValueError(
                    f"Unexpected operation '{node.name}' in circuit given to IQMOptimizeSingleQubitGates pass"
                )

        if not self._drop_final_rz:
            for qubit_index, rz_angle in enumerate(rz_angles):
                if rz_angle != 0:
                    qubit = dag.qubits[qubit_index]
                    dag.apply_operation_back(RGate(-np.pi, 0), qargs=(qubit,))
                    dag.apply_operation_back(RGate(np.pi, rz_angle / 2), qargs=(qubit,))

        return dag

    def _validate_ops(self, dag: DAGCircuit):
        valid_ops = self._basis + ["measure", "reset", "delay", "barrier"]
        for node in dag.op_nodes():
            if node.name not in valid_ops:
                raise ValueError(
                    f"Invalid operation '{node.name}' found in IQMOptimize1QbDecomposition pass, "
                    + f"expected operations {valid_ops}"
                )


def optimize_single_qubit_gates(
    circuit: QuantumCircuit, drop_final_rz: bool = True, ignore_barriers: bool = False
) -> QuantumCircuit:
    """Optimize number of single-qubit gates in a transpiled circuit exploiting the IQM specific gate set.

    Args:
        circuit: quantum circuit to optimize
        drop_final_rz: Drop terminal RZ gates even if there are no measurements following them (since they do not affect
            the measurement results). Note that this will change the unitary propagator of the circuit.
            It is recommended always to set this to true as the final RZ gates do no change the measurement outcomes of
            the circuit.
        ignore_barriers (bool): Removes barriers from the circuit if they exist (default = False) before optimization.

    Returns:
        optimized circuit

    """
    warnings.warn(
        DeprecationWarning(
            "This function is deprecated and will be removed in a later version of `iqm.qiskit_iqm`. "
            + "Single qubit gate optimization is now automatically applied when running `qiskit.transpile()` on any "
            + "IQM device. If you want to have more fine grained control over the optimization, please use the "
            + "`iqm.qiskit_iqm.transpile_to_IQM` function."
        )
    )
    # Code not updated to use transpile_to_IQM due to circular imports
    new_circuit = PassManager(IQMOptimizeSingleQubitGates(drop_final_rz, ignore_barriers)).run(circuit)
    new_circuit._layout = circuit.layout
    return new_circuit


class IQMReplaceGateWithUnitaryPass(TransformationPass):
    """Transpiler pass that replaces all gates with given name in a circuit with a UnitaryGate.

    Args:
        gate: The name of the gate to replace.
        unitary: The unitary matrix to replace the gate with.

    """

    def __init__(self, gate: str, unitary: list[list[float]]):
        super().__init__()
        self.gate = gate
        self.unitary = unitary

    def run(self, dag):
        for node in dag.op_nodes():
            if node.name == self.gate:
                dag.substitute_node(node, UnitaryGate(self.unitary))
        return dag



# Copyright 2022-2025 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fake backend for IQM's 6-qubit Deneb architecture."""

from iqm.iqm_client import StaticQuantumArchitecture
from iqm.qiskit_iqm.fake_backends.iqm_fake_backend import IQMErrorProfile, IQMFakeBackend


def IQMFakeDeneb() -> IQMFakeBackend:
    """Return IQMFakeBackend instance representing IQM's Deneb architecture."""
    architecture = StaticQuantumArchitecture(
        qubits=["QB1", "QB2", "QB3", "QB4", "QB5", "QB6"],
        computational_resonators=["CR1"],
        connectivity=[
            ("CR1", "QB1"),
            ("CR1", "QB2"),
            ("CR1", "QB3"),
            ("CR1", "QB4"),
            ("CR1", "QB5"),
            ("CR1", "QB6"),
        ],
    )
    error_profile = IQMErrorProfile(
        t1s={
            "CR1": 5400.0,
            "QB1": 35000.0,
            "QB2": 35000.0,
            "QB3": 35000.0,
            "QB4": 35000.0,
            "QB5": 35000.0,
            "QB6": 35000.0,
        },
        t2s={
            "CR1": 10800.0,
            "QB1": 33000.0,
            "QB2": 33000.0,
            "QB3": 33000.0,
            "QB4": 33000.0,
            "QB5": 33000.0,
            "QB6": 33000.0,
        },
        single_qubit_gate_depolarizing_error_parameters={
            "prx": {
                "QB1": 0.0002,
                "QB2": 0.0002,
                "QB3": 0.0002,
                "QB4": 0.0002,
                "QB5": 0.0002,
                "QB6": 0.0002,
            }
        },
        two_qubit_gate_depolarizing_error_parameters={
            "cz": {
                ("QB1", "CR1"): 0.0128,
                ("QB2", "CR1"): 0.0128,
                ("QB3", "CR1"): 0.0128,
                ("QB4", "CR1"): 0.0128,
                ("QB5", "CR1"): 0.0128,
                ("QB6", "CR1"): 0.0128,
            },
            "move": {
                ("QB1", "CR1"): 0.0,
                ("QB2", "CR1"): 0.0,
                ("QB3", "CR1"): 0.0,
                ("QB4", "CR1"): 0.0,
                ("QB5", "CR1"): 0.0,
                ("QB6", "CR1"): 0.0,
            },
        },
        single_qubit_gate_durations={"prx": 40.0},
        two_qubit_gate_durations={"cz": 120.0, "move": 96.0},
        readout_errors={
            "QB1": {"0": 0.02, "1": 0.02},
            "QB2": {"0": 0.02, "1": 0.02},
            "QB3": {"0": 0.02, "1": 0.02},
            "QB4": {"0": 0.02, "1": 0.02},
            "QB5": {"0": 0.02, "1": 0.02},
            "QB6": {"0": 0.02, "1": 0.02},
        },
        name="Deneb",
    )

    return IQMFakeBackend(architecture, error_profile, name="IQMFakeDenebBackend")



# Copyright 2022-2025 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Circuit execution jobs."""

from __future__ import annotations

from collections import Counter
from datetime import date
from typing import TYPE_CHECKING, Optional, Union
import uuid
import warnings

from iqm.iqm_client import (
    DEFAULT_TIMEOUT_SECONDS,
    CircuitMeasurementResults,
    HeraldingMode,
    IQMClient,
    JobAbortionError,
    RunRequest,
    RunResult,
    Status,
)
from iqm.qiskit_iqm.qiskit_to_iqm import MeasurementKey
import numpy as np
from qiskit.providers import JobStatus, JobV1
from qiskit.result import Counts, Result

if TYPE_CHECKING:
    from iqm.qiskit_iqm.iqm_provider import IQMBackend


class IQMJob(JobV1):
    """Implementation of Qiskit's job interface to handle circuit execution on an IQM server.

    Args:
        backend: Backend instance initiating this job.
        job_id: String representation of the UUID generated by IQM server.
        timeout_seconds: Maximum time to wait for the job to finish. By default, we use the
            :class:`~iqm.iqm_client.iqm_client.IQMClient` default.
        kwargs: Arguments to be passed to the initializer of the parent class.

    """

    def __init__(self, backend: IQMBackend, job_id: str, timeout_seconds: Optional[float] = None, **kwargs):
        super().__init__(backend, job_id=job_id, **kwargs)
        self._result: Union[None, list[tuple[str, list[str]]]] = None
        self._calibration_set_id: Optional[uuid.UUID] = None
        self._request: Optional[RunRequest] = None
        self._client: IQMClient = backend.client
        self.circuit_metadata: Optional[list] = None  # Metadata that was originally associated with circuits by user
        self._timeout_seconds: float = timeout_seconds if timeout_seconds is not None else DEFAULT_TIMEOUT_SECONDS

    def _format_iqm_results(self, iqm_result: RunResult) -> list[tuple[str, list[str]]]:
        """Convert the measurement results for a batch of circuits into the Qiskit format.

        Args:
            iqm_result: measurement results for the circuit batch
        Returns:
            A list of (circuit_name, measurements) tuples, one tuple for each circuit in the batch.
            The measurements are a list of bitstrings, one per shot, representing the state of the classical
            registers after the shot.

        """
        if iqm_result.measurements is None:
            raise ValueError(
                f'Cannot format IQM result without measurements. Job status is "{iqm_result.status.value.upper()}"'
            )
        requested_shots = self.metadata.get("shots", iqm_result.metadata.shots)
        # If no heralding, for all circuits we expect the same number of shots which is the shots requested by user.
        expect_exact_shots = iqm_result.metadata.heralding_mode == HeraldingMode.NONE

        return [
            (circuit.name, self._format_measurement_results(measurements, requested_shots, expect_exact_shots))
            for measurements, circuit in zip(iqm_result.measurements, iqm_result.metadata.circuits)
        ]

    @staticmethod
    def _format_measurement_results(
        measurement_results: CircuitMeasurementResults, requested_shots: int, expect_exact_shots: bool = True
    ) -> list[str]:
        """Convert the measurement results from a circuit into the Qiskit format.

        Args:
            measurement_results: measurement results for a single circuit
            requested_shots: number of shots requested
            expect_exact_shots: iff True, we must get exactly as many shots as requested
        Returns:
            For each shot, a bitstring representing the state of the classical registers after the
            shot, in little-endian order.

        """
        # Mapping from creg index (in the circuit) to an array with shape (shots, len(creg)) with the results.
        formatted_results: dict[int, np.ndarray] = {}
        for k, v in measurement_results.items():
            # measurement keys encode data about the classical registers in the original Qiskit circuit
            mk = MeasurementKey.from_string(k)
            res = np.array(v, dtype=int)
            shots = len(res)
            if shots == 0 and not expect_exact_shots:
                warnings.warn(
                    "Received measurement results containing zero shots. "
                    "In case you are using non-default heralding mode, this could be because of bad calibration."
                )
                res = np.array([])
            else:
                # in Qiskit each measurement is a separate single-qubit instruction. qiskit-iqm assigns unique
                # measurement key to each such instruction, so only one column is expected per measurement key.
                if res.shape[1] != 1:
                    raise ValueError(f"Measurement result {mk} has the wrong shape {res.shape}, expected (*, 1)")
                res = res[:, 0]

            if expect_exact_shots and shots != requested_shots:
                raise ValueError(f"Expected {requested_shots} shots but got {shots} for measurement result {mk}")

            # group the measurements into cregs, fill in zeros for unused bits
            creg = formatted_results.setdefault(mk.creg_idx, np.zeros((shots, mk.creg_len), dtype=int))
            creg[:, mk.clbit_idx] = res

        # TODO If the original circuit has a creg that is not used at all we won't know about it here,
        # and thus cannot include it (containing only zeros) in the result strings.

        # Number of shots is the same for all measurement keys.
        # Qiskit uses the little-endian convention in presenting the result bitstrings
        # (both between and within registers), hence the [::-1]
        return [
            " ".join("".join(map(str, res[s, :])) for _, res in sorted(formatted_results.items()))[::-1]
            for s in range(shots)
        ]

    def submit(self):
        raise NotImplementedError(
            "You should never have to submit jobs by calling this method. When running circuits through "
            "RemoteIQMBackend, the submission will happen under the hood. The job instance that you get is only for "
            "checking the progress and retrieving the results of the submitted job."
        )

    def cancel(self) -> bool:
        """Attempt to cancel the job.

        Returns:
            True if the job was cancelled successfully, False otherwise

        """
        try:
            self._client.abort_job(uuid.UUID(self._job_id))
            return True
        except JobAbortionError as e:
            warnings.warn(f"Failed to cancel job: {e}")
            return False

    def result(self) -> Result:
        if not self._result:
            # client will raise an error if it was unable to get the results within the timeout
            results = self._client.wait_for_results(uuid.UUID(self._job_id), self._timeout_seconds)
            self._calibration_set_id = results.metadata.calibration_set_id
            self._request = results.metadata.request
            if results.metadata.timestamps is not None:
                self.metadata["timestamps"] = results.metadata.timestamps.copy()
            self._result = self._format_iqm_results(results)
            # IQMBackend.run() populates IQMJob.circuit_metadata, so it may be None if this IQMJob
            # was created manually from a job_id. In that case retrieve circuit metadata from
            # RunResult.metadata.request.circuits[n].metadata
            if self.circuit_metadata is None and results.metadata.request is not None:
                self.circuit_metadata = [c.metadata for c in results.metadata.circuits]

        result_dict = {
            "backend_name": None,
            "backend_version": None,
            "qobj_id": None,
            "job_id": self._job_id,
            "success": True,
            "results": [
                {
                    "shots": len(measurement_results),
                    "success": True,
                    "data": {
                        "memory": measurement_results,
                        "counts": Counts(Counter(measurement_results)),
                        "metadata": self.circuit_metadata[i] if self.circuit_metadata is not None else {},
                    },
                    "header": {"name": name},
                    "calibration_set_id": self._calibration_set_id,
                }
                for i, (name, measurement_results) in enumerate(self._result)
            ],
            "date": date.today().isoformat(),
            "request": self._request,
            "timestamps": self.metadata.get("timestamps"),
        }
        return Result.from_dict(result_dict)

    def status(self) -> JobStatus:
        if self._result:
            return JobStatus.DONE

        result = self._client.get_run_status(uuid.UUID(self._job_id))
        if result.status == Status.PENDING_EXECUTION:
            return JobStatus.RUNNING
        if result.status == Status.READY:
            return JobStatus.DONE
        if result.status in {Status.FAILED, Status.DELETION_FAILED}:
            return JobStatus.ERROR
        if result.status in {Status.ABORTED, Status.DELETED}:
            return JobStatus.CANCELLED
        # For everything else, we assume that it's in progress one way or another
        return JobStatus.QUEUED

    def queue_position(self, refresh: bool = False) -> Optional[int]:
        """Return the position of the job in the server queue.

        Note:
            The position is not yet implemented and this function will always
            return ``None``. The ``refresh`` argument is ignored.

        Args:
            refresh: If ``True``, re-query the server to get the latest value.
                Otherwise return the cached value.

        Returns:
            Position in the queue or ``None`` if position is unknown or not applicable.

        """
        return None

    def error_message(self) -> Optional[str]:
        """Returns the error message if job has failed, otherwise returns None."""
        return self._client.get_run_status(uuid.UUID(self._job_id)).message



# Copyright 2024 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions for circuit validation."""

from typing import Optional

from iqm.iqm_client import Circuit as IQMClientCircuit
from iqm.iqm_client import IQMClient, MoveGateValidationMode
from iqm.qiskit_iqm.iqm_backend import IQMBackendBase
from iqm.qiskit_iqm.qiskit_to_iqm import serialize_instructions
from qiskit import QuantumCircuit


def validate_circuit(
    circuit: QuantumCircuit,
    backend: IQMBackendBase,
    validate_moves: Optional[MoveGateValidationMode] = None,
    qubit_mapping: Optional[dict[int, str]] = None,
):
    """Validate a circuit against the backend."""
    if qubit_mapping is None:
        qubit_mapping = backend._idx_to_qb
    new_circuit = IQMClientCircuit(
        name="Validation circuit",
        instructions=tuple(serialize_instructions(circuit=circuit, qubit_index_to_name=qubit_mapping)),
    )
    if validate_moves is None:
        validate_moves = MoveGateValidationMode.STRICT
    IQMClient._validate_circuit_instructions(
        architecture=backend.architecture,
        circuits=[new_circuit],
        validate_moves=validate_moves,
    )



# Copyright 2024 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Collection of Qiskit transpiler plugins for native use of specialized transpiler passes by our devices."""

from typing import Optional

from iqm.iqm_client.transpile import ExistingMoveHandlingOptions
from iqm.qiskit_iqm.iqm_backend import IQMTarget
from iqm.qiskit_iqm.iqm_naive_move_pass import IQMNaiveResonatorMoving
from iqm.qiskit_iqm.iqm_transpilation import IQMOptimizeSingleQubitGates
from qiskit.transpiler.passmanager import PassManager
from qiskit.transpiler.passmanager_config import PassManagerConfig
from qiskit.transpiler.preset_passmanagers.builtin_plugins import PassManagerStagePlugin


class IQMSchedulingPlugin(PassManagerStagePlugin):
    """Basic plugin for scheduling stage of IQM devices.

    Args:
        move_gate_routing: whether to include MoveGate routing in the scheduling stage.
        optimize_sqg: Whether to include single qubit gate optimization in the scheduling stage.
        drop_final_rz: Whether to drop trailing RZ gates in the circuit during single qubit gate optimization.
        ignore_barriers: Whether to ignore barriers during single qubit gate optimization.
        existing_move_handling: How to handle existing MoveGates in the circuit during MoveGate routing.

    Raises:
        ValueError: When incompatible options are set.

    """

    def __init__(
        self,
        move_gate_routing: bool,
        optimize_sqg: bool,
        drop_final_rz: bool,
        ignore_barriers: bool,
        existing_move_handling: Optional[ExistingMoveHandlingOptions],
    ) -> None:
        super().__init__()
        self.move_gate_routing = move_gate_routing
        self.optimize_sqg = optimize_sqg
        self.drop_final_rz = drop_final_rz
        self.ignore_barriers = ignore_barriers
        self.existing_move_handling = (
            existing_move_handling if existing_move_handling is not None else ExistingMoveHandlingOptions.KEEP
        )

    def pass_manager(
        self, pass_manager_config: PassManagerConfig, optimization_level: Optional[int] = None
    ) -> PassManager:
        """Build scheduling stage PassManager"""
        scheduling = PassManager()
        if self.optimize_sqg:
            scheduling.append(
                IQMOptimizeSingleQubitGates(drop_final_rz=self.drop_final_rz, ignore_barriers=self.ignore_barriers)
            )
        if pass_manager_config.target is None:
            raise ValueError("PassManagerConfig must have a target backend set, unable to schedule MoveGate routing.")
        if self.move_gate_routing and isinstance(pass_manager_config.target, IQMTarget):
            scheduling.append(
                IQMNaiveResonatorMoving(
                    target=pass_manager_config.target,
                    existing_moves_handling=self.existing_move_handling,
                )
            )
        return scheduling


class MoveGateRoutingPlugin(IQMSchedulingPlugin):
    """Plugin class for IQM single qubit gate optimization and MoveGate routing as a scheduling stage."""

    def __init__(
        self,
        optimize_sqg: bool = True,
        drop_final_rz: bool = True,
        ignore_barriers: bool = False,
        existing_move_handling: Optional[ExistingMoveHandlingOptions] = None,
    ) -> None:
        super().__init__(True, optimize_sqg, drop_final_rz, ignore_barriers, existing_move_handling)


class MoveGateRoutingOnlyPlugin(MoveGateRoutingPlugin):
    """Plugin class for MoveGate routing without single qubit gate optimization as a scheduling stage."""

    def __init__(self):
        super().__init__(optimize_sqg=False)


class MoveGateRoutingKeepExistingMovesPlugin(MoveGateRoutingPlugin):
    """Plugin class for single qubit gate optimization and MoveGate routing where existing moves are kept."""

    def __init__(self):
        super().__init__(
            optimize_sqg=True,
            existing_move_handling=ExistingMoveHandlingOptions.KEEP,
        )


class MoveGateRoutingRemoveExistingMovesPlugin(MoveGateRoutingPlugin):
    """Plugin class for single qubit gate optimization and MoveGate routing where existing moves are removed."""

    def __init__(self):
        super().__init__(
            optimize_sqg=True,
            existing_move_handling=ExistingMoveHandlingOptions.REMOVE,
        )


class MoveGateRoutingTrustExistingMovesPlugin(MoveGateRoutingPlugin):
    """Plugin class for single qubit gate optimization and MoveGate routing where existing moves are not checked."""

    def __init__(self):
        super().__init__(
            optimize_sqg=True,
            existing_move_handling=ExistingMoveHandlingOptions.TRUST,
        )


class MoveGateRoutingWithExactRZPlugin(MoveGateRoutingPlugin):
    """Plugin class for single qubit gate optimization and MoveGate routing where
    trailing RZ gates are kept in the circuit.
    """

    def __init__(self):
        super().__init__(optimize_sqg=True, drop_final_rz=False)


class MoveGateRoutingWithRZOptimizationIgnoreBarriersPlugin(MoveGateRoutingPlugin):
    """Plugin class for single qubit gate optimization and MoveGate routing where barriers are ignored during
    optimization.
    """

    def __init__(self):
        super().__init__(
            optimize_sqg=True,
            ignore_barriers=True,
        )


class MoveGateRoutingOnlyKeepExistingMovesPlugin(MoveGateRoutingPlugin):
    """Plugin class for MoveGate routing without single qubit gate optimization
    where existing moves are kept.
    """

    def __init__(self):
        super().__init__(
            optimize_sqg=False,
            existing_move_handling=ExistingMoveHandlingOptions.KEEP,
        )


class MoveGateRoutingOnlyRemoveExistingMovesPlugin(MoveGateRoutingPlugin):
    """Plugin class for MoveGate routing without single qubit gate optimization
    where existing moves are removed.
    """

    def __init__(self):
        super().__init__(
            optimize_sqg=False,
            existing_move_handling=ExistingMoveHandlingOptions.REMOVE,
        )


class MoveGateRoutingOnlyTrustExistingMovesPlugin(MoveGateRoutingPlugin):
    """Plugin class for MoveGate routing without single qubit gate optimization
    where existing moves are not checked.
    """

    def __init__(self):
        super().__init__(
            optimize_sqg=False,
            existing_move_handling=ExistingMoveHandlingOptions.TRUST,
        )


class OnlyRZOptimizationPlugin(IQMSchedulingPlugin):
    """Plugin class for single qubit gate optimization without MOVE gate routing."""

    def __init__(
        self,
        drop_final_rz=True,
        ignore_barriers=False,
    ):
        super().__init__(False, True, drop_final_rz, ignore_barriers, None)


class OnlyRZOptimizationExactPlugin(OnlyRZOptimizationPlugin):
    """Plugin class for single qubit gate optimization without MOVE gate routing and
    the final RZ gates are not dropped.
    """

    def __init__(self):
        super().__init__(drop_final_rz=False)


class OnlyRZOptimizationIgnoreBarriersPlugin(OnlyRZOptimizationPlugin):
    """Plugin class for single qubit gate optimization without MOVE gate routing where barriers are ignored."""

    def __init__(self):
        super().__init__(ignore_barriers=True)


class OnlyRZOptimizationExactIgnoreBarriersPlugin(OnlyRZOptimizationPlugin):
    """Plugin class for single qubit gate optimization without MOVE gate routing and
    the final RZ gates are not dropped.
    """

    def __init__(self):
        super().__init__(drop_final_rz=False, ignore_barriers=True)


class IQMDefaultSchedulingPlugin(IQMSchedulingPlugin):
    """Plugin class for IQM single qubit gate optimization and MoveGate routing as a scheduling stage."""

    def __init__(self) -> None:
        super().__init__(
            True, optimize_sqg=True, drop_final_rz=True, ignore_barriers=False, existing_move_handling=None
        )

    def pass_manager(
        self, pass_manager_config: PassManagerConfig, optimization_level: Optional[int] = None
    ) -> PassManager:
        """Build scheduling stage PassManager"""
        if optimization_level == 0:
            self.optimize_sqg = False
        return super().pass_manager(pass_manager_config, optimization_level)



# Copyright 2022-2023 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fake backend for IQM's 5-qubit Adonis architecture."""

from iqm.iqm_client import StaticQuantumArchitecture
from iqm.qiskit_iqm.fake_backends.iqm_fake_backend import IQMErrorProfile, IQMFakeBackend


def IQMFakeAdonis() -> IQMFakeBackend:
    """Return IQMFakeBackend instance representing IQM's Adonis architecture."""
    architecture = StaticQuantumArchitecture(
        qubits=["QB1", "QB2", "QB3", "QB4", "QB5"],
        computational_resonators=[],
        connectivity=[("QB1", "QB3"), ("QB2", "QB3"), ("QB3", "QB4"), ("QB3", "QB5")],
    )
    error_profile = IQMErrorProfile(
        t1s={"QB1": 27000.0, "QB2": 33000.0, "QB3": 25000.0, "QB4": 40000.0, "QB5": 25000.0},
        t2s={"QB1": 20000.0, "QB2": 26000.0, "QB3": 23000.0, "QB4": 26000.0, "QB5": 7000.0},
        single_qubit_gate_depolarizing_error_parameters={
            "prx": {"QB1": 0.0006, "QB2": 0.0054, "QB3": 0.0001, "QB4": 0.0, "QB5": 0.0005}
        },
        two_qubit_gate_depolarizing_error_parameters={
            "cz": {("QB1", "QB3"): 0.0335, ("QB2", "QB3"): 0.0344, ("QB3", "QB4"): 0.0192, ("QB3", "QB5"): 0.0373}
        },
        single_qubit_gate_durations={"prx": 40.0},
        two_qubit_gate_durations={"cz": 80.0},
        readout_errors={
            "QB1": {"0": 0.021, "1": 0.021},
            "QB2": {"0": 0.018, "1": 0.018},
            "QB3": {"0": 0.056, "1": 0.056},
            "QB4": {"0": 0.021, "1": 0.021},
            "QB5": {"0": 0.023, "1": 0.023},
        },
        name="sample-chip",
    )

    return IQMFakeBackend(architecture, error_profile, name="IQMFakeAdonisBackend")



# Copyright 2024 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Naive transpilation for the IQM Star architecture."""

from typing import Optional, Union
import warnings

from iqm.iqm_client import Circuit as IQMClientCircuit
from iqm.iqm_client.transpile import ExistingMoveHandlingOptions, transpile_insert_moves
import numpy as np
from pydantic_core import ValidationError
from qiskit import QuantumCircuit, transpile
from qiskit.circuit.library import RGate
from qiskit.converters import circuit_to_dag, dag_to_circuit
from qiskit.dagcircuit import DAGCircuit
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.layout import Layout

from .iqm_backend import IQMBackendBase, IQMTarget
from .iqm_move_layout import generate_initial_layout
from .qiskit_to_iqm import deserialize_instructions, serialize_instructions


class IQMNaiveResonatorMoving(TransformationPass):
    """Naive transpilation pass for resonator moving.

    The logic of this pass is deferred to `iqm-client.transpile_insert_moves`.
    This pass is a wrapper that converts the circuit into the IQMClient Circuit format,
    runs the `transpile_insert_moves` function, and then converts the result back to a Qiskit circuit.

    Args:
        target: Transpilation target.
        existing_moves_handling: How to handle existing MOVE gates in the circuit.

    """

    def __init__(
        self,
        target: IQMTarget,
        existing_moves_handling: ExistingMoveHandlingOptions = ExistingMoveHandlingOptions.KEEP,
    ):
        super().__init__()
        self.target = target
        self.architecture = target.iqm_dqa
        self.idx_to_component = target.iqm_idx_to_component
        self.component_to_idx = target.iqm_component_to_idx
        self.existing_moves_handling = existing_moves_handling

    def run(self, dag: DAGCircuit) -> DAGCircuit:
        """Run the pass on a circuit.

        Args:
            dag: DAG to map.

        Returns:
            Mapped ``dag``.

        Raises:
            TranspilerError: The layout is not compatible with the DAG, or if the input gate set is incorrect.

        """
        # TODO: Temporary hack to get the symbolic parameters to work: replace symbols with (inf, idx).
        # Replace symbolic parameters with indices and store the index to symbol mapping.
        symbolic_gates = {}
        symbolic_index = 0
        for node in dag.topological_op_nodes():
            # This only works for prx gates because that has two parameters
            # We use one to mark that it is a symbolic gate (np.inf) and the other to store the index.
            if node.name == "r" and not all(isinstance(param, (float, int)) for param in node.op.params):
                symbolic_gates[symbolic_index] = node.op.params
                dag.substitute_node(node, RGate(np.inf, float(symbolic_index)))
                symbolic_index += 1

        circuit = dag_to_circuit(dag)
        if len(circuit) == 0:
            return dag  # Empty circuit, no need to transpile.
        # For some reason, the dag does not contain the layout, so we need to do a bunch of fixing.
        if self.property_set.get("layout"):
            layout = self.property_set["layout"]
        else:
            # Reconstruct the layout from the dag.
            layout = Layout()
            for qreg in dag.qregs:
                layout.add_register(qreg)
            for i, qubit in enumerate(dag.qubits):
                layout.add(qubit, i)

        # Convert the circuit to the IQMClientCircuit format and run the transpiler.
        iqm_circuit = IQMClientCircuit(
            name="Transpiling Circuit",
            instructions=tuple(serialize_instructions(circuit, self.idx_to_component)),
        )
        try:
            routed_iqm_circuit = transpile_insert_moves(
                iqm_circuit, self.architecture, existing_moves=self.existing_moves_handling
            )
            routed_circuit = deserialize_instructions(
                list(routed_iqm_circuit.instructions), self.component_to_idx, layout
            )
        except ValidationError as e:  # The Circuit without move gates is empty.
            errors = e.errors()
            if (
                len(errors) == 1
                and errors[0]["msg"] == "Value error, Each circuit should have at least one instruction."
            ):
                circ_args = [circuit.num_ancillas, circuit.num_clbits]
                routed_circuit = QuantumCircuit(*layout.get_registers(), *(arg for arg in circ_args if arg > 0))
            else:
                raise e

        # Create the new DAG and make sure that the qubits are properly ordered.
        ordered_qubits = [layout.get_physical_bits()[i] for i in range(len(layout.get_physical_bits()))]
        new_dag = circuit_to_dag(routed_circuit, qubit_order=ordered_qubits, clbit_order=routed_circuit.clbits)

        # Reinsert the symbolic parameters.
        for node in new_dag.topological_op_nodes():
            # This only works for prx gates because that has two parameters
            # We use one to mark that it is a symbolic gate (np.inf) and the other to store the index.
            if node.name == "r" and not np.isfinite(node.op.params[0]):
                new_dag.substitute_node(node, RGate(*symbolic_gates[int(node.op.params[1])]))

        # Update the final_layout with the correct bits.
        if "final_layout" in self.property_set:
            inv_layout = layout.get_physical_bits()
            new_final_layout_dict = {
                physical: inv_layout[dag.find_bit(virtual).index]
                for physical, virtual in self.property_set["final_layout"].get_physical_bits().items()
            }
            resonator_dict = {
                phys: inv_layout[new_dag.find_bit(virt).index]
                for phys, virt in inv_layout.items()
                if phys not in self.property_set["final_layout"].get_physical_bits()
            }
            new_final_layout_dict.update(resonator_dict)
            self.property_set["final_layout"] = Layout(new_final_layout_dict)
        else:
            self.property_set["final_layout"] = layout

        return new_dag


def _get_scheduling_method(
    perform_move_routing: bool,
    optimize_single_qubits: bool,
    remove_final_rzs: bool,
    ignore_barriers: bool,
    existing_moves_handling: Optional[ExistingMoveHandlingOptions],
) -> str:
    """Determine scheduling based on flags."""
    if perform_move_routing:
        if optimize_single_qubits:
            if not remove_final_rzs and ignore_barriers and existing_moves_handling is None:
                raise ValueError(
                    f"Move gate routing not compatible with {optimize_single_qubits=}, "
                    f"{remove_final_rzs=}, and {ignore_barriers=}."
                )
            if not remove_final_rzs:
                scheduling_method = "move_routing_exact_global_phase"
            elif ignore_barriers:
                scheduling_method = "move_routing_rz_optimization_ignores_barriers"
            else:
                scheduling_method = "move_routing"
        else:
            scheduling_method = "only_move_routing"
        if existing_moves_handling is not None:
            if not scheduling_method.endswith("routing"):
                raise ValueError(
                    "Existing Move handling options are not compatible with `remove_final_rzs` and \
                    `ignore_barriers` options."
                )  # No technical reason for this, just hard to maintain all combinations.
            scheduling_method += "_" + existing_moves_handling.value
    elif optimize_single_qubits:
        scheduling_method = "only_rz_optimization"
        if not remove_final_rzs:
            scheduling_method += "_exact_global_phase"
        if ignore_barriers:
            scheduling_method += "_ignore_barriers"
    else:
        scheduling_method = "default"
    return scheduling_method


def transpile_to_IQM(  # noqa: PLR0913
    circuit: QuantumCircuit,
    backend: IQMBackendBase,
    target: Optional[IQMTarget] = None,
    initial_layout: Optional[Union[Layout, dict, list]] = None,
    perform_move_routing: bool = True,
    optimize_single_qubits: bool = True,
    ignore_barriers: bool = False,
    remove_final_rzs: bool = True,
    existing_moves_handling: Optional[ExistingMoveHandlingOptions] = None,
    restrict_to_qubits: Optional[Union[list[int], list[str]]] = None,
    **qiskit_transpiler_kwargs,
) -> QuantumCircuit:
    """Customized transpilation to IQM backends.

    Works with both the Crystal and Star architectures.

    Args:
        circuit: The circuit to be transpiled without MOVE gates.
        backend: The target backend to compile to. Does not require a resonator.
        target: An alternative target to compile to than the backend, using this option requires intimate knowledge
            of the transpiler and thus it is not recommended to use.
        initial_layout: The initial layout to use for the transpilation, same as :func:`~qiskit.compiler.transpile`.
        perform_move_routing: Whether to perform MOVE gate routing.
        optimize_single_qubits: Whether to optimize single qubit gates away.
        ignore_barriers: Whether to ignore barriers when optimizing single qubit gates away.
        remove_final_rzs: Whether to remove the final z rotations. It is recommended always to set this to true as
            the final RZ gates do no change the measurement outcomes of the circuit.
        existing_moves_handling: How to handle existing MOVE gates in the circuit, required if the circuit contains
            MOVE gates.
        restrict_to_qubits: Restrict the transpilation to only use these specific physical qubits. Note that you will
            have to pass this information to the ``backend.run`` method as well as a dictionary.
        qiskit_transpiler_kwargs: Arguments to be passed to the Qiskit transpiler.

    Returns:
        Transpiled circuit ready for running on the backend.

    """
    if restrict_to_qubits is not None:
        restrict_to_qubits = [backend.qubit_name_to_index(q) if isinstance(q, str) else q for q in restrict_to_qubits]

    if target is None:
        if circuit.count_ops().get("move", 0) > 0:
            target = backend.target_with_resonators
            # Create a sensible initial layout if none is provided
            if initial_layout is None:
                initial_layout = generate_initial_layout(backend, circuit, restrict_to_qubits)
            if perform_move_routing and existing_moves_handling is None:
                raise ValueError("The circuit contains MOVE gates but existing_moves_handling is not set.")
        else:
            target = backend.target

    if restrict_to_qubits is not None:
        target = target.restrict_to_qubits(restrict_to_qubits)

    # Determine which scheduling method to use
    scheduling_method = qiskit_transpiler_kwargs.pop("scheduling_method", None)
    if scheduling_method is None:
        scheduling_method = _get_scheduling_method(
            perform_move_routing=perform_move_routing,
            optimize_single_qubits=optimize_single_qubits,
            remove_final_rzs=remove_final_rzs,
            ignore_barriers=ignore_barriers,
            existing_moves_handling=existing_moves_handling,
        )
    else:
        warnings.warn(
            f"Scheduling method is set to {scheduling_method}, but it is normally used to pass other transpiler "
            + "options, ignoring the `perform_move_routing`, `optimize_single_qubits`, `remove_final_rzs`, "
            + "`ignore_barriers`, and `existing_moves_handling` arguments."
        )
    qiskit_transpiler_kwargs["scheduling_method"] = scheduling_method
    new_circuit = transpile(circuit, target=target, initial_layout=initial_layout, **qiskit_transpiler_kwargs)
    return new_circuit



# Copyright 2024-2025 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generate an initial layout for a quantum circuit that is
valid on the quantum architecture specification of the given backend.
"""

from typing import Optional, Union

from iqm.qiskit_iqm.iqm_backend import IQMBackendBase
from qiskit import QuantumCircuit
from qiskit.circuit import Qubit
from qiskit.dagcircuit import DAGCircuit
from qiskit.transpiler import PassManager, TranspilerError
from qiskit.transpiler.layout import Layout
from qiskit.transpiler.passes import TrivialLayout


class IQMMoveLayout(TrivialLayout):
    """Create a layout that is valid on the dynamic quantum architecture of the
    given IQM target.

    The architecture defines which gate loci are available. This class
    tries to map the virtual/logical components of the circuit to the physical QPU components
    of the architecture, such that the gates in the circuit can be applied on those components.

    This class is required because Qiskit's basic layout algorithm assumes all connections between
    two qubits have the same two-qubit gates available, which isn't true in general.

    .. note::

       This version of the layout generator only works reliably with a single resonator,
       and can only handle pure Star architecture circuits.
       It also assumes that a valid layout exists for the circuit that does not require SWAPs, which
       isn't true in general.
    """

    def run(self, dag: DAGCircuit):
        """Creates a valid layout for the given quantum circuit.

        Args:
            dag: Circuit DAG to find layout for.

        Raises:
            TranspilerError: A valid layout could not be found.

        """
        target = self.target

        # NOTE assumes we use the real Star architecture here
        reqs, resonators = self._calculate_requirements(dag)
        if len(resonators) > 1:
            raise TranspilerError(
                "Circuit requires more than one computational resonator, IQMMoveLayout cannot yet handle this."
            )

        # map unused physical components to the gates they support
        free_qubits: dict[int, set[str]] = {}
        for op in ["move", "cz", "measure", "r"]:
            for locus in target.qargs_for_operation_name(op):
                # For arity-2 gates, we only care about the first locus component,
                # because in the Star architecture the second one is a resonator.
                free_qubits.setdefault(locus[0], set()).add(op)

        # Add reqs for unused logical components in the circuit (they require nothing) so they too end up in the layout.
        # They are assigned physical components only after all the used components have been assigned theirs, hence
        # they will not affect the satisfiability.
        for log_idx in range(len(dag.qubits)):
            if log_idx not in resonators:
                reqs.setdefault(log_idx, set())

        # mapping from physical component index to logical qubit
        layout: dict[int, Qubit] = {}
        # add the resonators to the layout
        # TODO if we require more than one resonator, the order in which we map them to physical resonators matters
        for log_idx, res_name in zip(resonators, target.iqm_dqa.computational_resonators):
            phys_idx = target.iqm_component_to_idx[res_name]
            layout = {phys_idx: dag.qubits[log_idx]}

        # add qubits to the layout
        for log_idx, req_ops in reqs.items():
            # which physical qubits have the required ops available?
            mapping_options = [(phys_idx, len(av_ops)) for phys_idx, av_ops in free_qubits.items() if req_ops <= av_ops]
            # pick the one that has the fewest unneeded ops
            # TODO this heuristic does not always find a possible layout! If there are two options, each with a
            # different unnecessary gate, we pick one essentially at random, which may mean that later we cannot
            # satisfy another logical component because we picked the wrong option.
            if not mapping_options:
                raise TranspilerError(
                    f"Cannot find a physical qubit to map logical qubit {log_idx} to, "
                    f"requires {req_ops}, available: {free_qubits}."
                )
            phys_idx, _ = min(mapping_options, key=lambda x: x[1])
            layout[phys_idx] = dag.qubits[log_idx]
            del free_qubits[phys_idx]  # phys_idx is now taken

        self.property_set["layout"] = Layout(layout)

    def get_initial_layout(self) -> Layout:
        """Returns the initial layout generated by the algorithm.

        Returns:
            The initial layout.

        """
        return self.property_set["layout"]

    @staticmethod
    def _calculate_requirements(dag: DAGCircuit) -> tuple[dict[int, set[str]], set[int]]:
        """Determine the requirements for each used logical qubit in the circuit.

        Because in the Star architecture two-qubit gates have (qubit, resonator) loci, based on them
        we can figure out which logical qubits must be mapped to computational resonators.

        Unused logical qubits do not appear in the circuit DAG, and thus are not included in the mapping.

        Args:
            dag: circuit to check

        Returns:
            Mapping of the logical qubit indices to the required gates for that qubit,
            logical qubit indices that must be resonators.

        """
        reqs: dict[int, set[str]] = {}
        resonators: set[int] = set()
        qubit_to_idx: dict[Qubit, int] = {qubit: log_idx for log_idx, qubit in enumerate(dag.qubits)}

        def _require_qubit_type(qubit: Qubit, required_type: str):
            """Add a requirement for the given qubit."""
            log_idx = qubit_to_idx[qubit]
            if log_idx in resonators:
                raise TranspilerError(
                    f"Virtual/logical qubit {qubit} for the '{node.name}' operation must be a qubit, "
                    f"but it is already required to be a resonator."
                )
            reqs.setdefault(log_idx, set()).add(required_type)

        for node in dag.topological_op_nodes():
            if node.name in ("barrier",):
                continue
            if node.name in ("move", "cz"):
                # In the real Star architecture, all arity-2 ops act on a (qubit, resonator) locus.
                qubit, resonator = node.qargs
                _require_qubit_type(qubit, node.name)
                # resonator
                log_idx = qubit_to_idx[resonator]
                if log_idx in reqs:
                    raise TranspilerError(
                        f"Virtual/logical qubit {resonator} for the '{node.name}' operation must be a resonator, "
                        f"but it is already required to be a qubit({reqs[log_idx]})."
                    )
                resonators.add(log_idx)
            elif node.name == "measure":
                _require_qubit_type(node.qargs[0], "measure")
            else:
                # all single-qubit gates are mapped to r
                _require_qubit_type(node.qargs[0], "r")

        return reqs, resonators


def generate_initial_layout(
    backend: IQMBackendBase,
    circuit: QuantumCircuit,
    restrict_to_qubits: Optional[Union[list[int], list[str]]] = None,
) -> Layout:
    """Generates an initial layout for the given circuit, when run against the given backend.

    Args:
        backend: IQM backend to run against.
        circuit: Star architecture circuit for which a layout is to be generated.
        restrict_to_qubits: Optional list of qubits to restrict the layout to.

    Returns:
        Layout that maps the logical qubits of ``circuit`` to the physical qubits of ``backend`` so that
        all the gates in ``circuit`` are available on those loci.

    """
    target = backend.get_real_target()
    if restrict_to_qubits is not None:
        target = target.restrict_to_qubits(restrict_to_qubits)

    layout_gen = IQMMoveLayout(target)
    pm = PassManager(layout_gen)
    pm.run(circuit)
    return layout_gen.get_initial_layout()



# Copyright 2020–2022 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for serializing and deserializing quantum circuits between Cirq and IQM Circuit formats."""

from cirq import Circuit, KeyCondition, NamedQid
from cirq.ops import (
    ClassicallyControlledOperation,
    CZPowGate,
    Gate,
    MeasurementGate,
    Operation,
    PhasedXPowGate,
    XPowGate,
    YPowGate,
)
from iqm import iqm_client
from iqm.cirq_iqm.iqm_gates import IQMMoveGate
from iqm.iqm_client import Instruction

# Mapping from IQM operation names to cirq operations
_IQM_CIRQ_OP_MAP: dict[str, tuple[type[Gate], ...]] = {
    # XPow and YPow kept for convenience, Cirq does not know how to decompose them into PhasedX
    # so we would have to add those rules...
    "prx": (PhasedXPowGate, XPowGate, YPowGate),
    "cz": (CZPowGate,),
    "move": (IQMMoveGate,),
    "measure": (MeasurementGate,),
    "cc_prx": (PhasedXPowGate, XPowGate, YPowGate),
}


def instruction_to_operation(instr: Instruction) -> Operation:
    """Convert an IQM instruction to a Cirq Operation.

    Args:
        instr: the IQM instruction

    Returns:
        Operation: the converted operation

    Raises:
        OperationNotSupportedError When the circuit contains an unsupported operation.

    """
    if instr.name not in _IQM_CIRQ_OP_MAP:
        raise OperationNotSupportedError(f"Operation {instr.name} not supported.")

    gate_type = _IQM_CIRQ_OP_MAP[instr.name][0]
    args_map = {"angle_t": "exponent", "phase_t": "phase_exponent"}
    args = {args_map[k]: v * 2 for k, v in instr.args.items() if k in args_map}
    qubits = [NamedQid(qubit, dimension=2) for qubit in instr.qubits]
    if instr.name == "measure":
        args["key"] = instr.args["key"]
        args["num_qubits"] = len(instr.qubits)
    elif instr.name == "cc_prx":
        return gate_type(**args)(*qubits).with_classical_controls(instr.args["feedback_key"])
    return gate_type(**args)(*qubits)


class OperationNotSupportedError(RuntimeError):
    """Raised when a given operation is not supported by the IQM server."""


def map_operation(operation: Operation) -> Instruction:
    """Map a Cirq Operation to the IQM data transfer format.

    Assumes the circuit has been transpiled so that it only contains operations natively supported by the
    given IQM quantum architecture.

    Args:
        operation: a Cirq Operation

    Returns:
        Instruction: the converted operation

    Raises:
        OperationNotSupportedError When the circuit contains an unsupported operation.

    """
    qubits = [qubit.name if isinstance(qubit, NamedQid) else str(qubit) for qubit in operation.qubits]
    if isinstance(operation.gate, (PhasedXPowGate, XPowGate, YPowGate)):
        return Instruction(
            name="prx",
            qubits=tuple(qubits),
            args={"angle_t": operation.gate.exponent / 2, "phase_t": operation.gate.phase_exponent / 2},
        )
    if isinstance(operation.gate, MeasurementGate):
        if any(operation.gate.full_invert_mask()):
            raise OperationNotSupportedError("Invert mask not supported")

        return Instruction(
            name="measure",
            qubits=tuple(qubits),
            args={"key": operation.gate.key},
        )
    if isinstance(operation.gate, CZPowGate):
        if operation.gate.exponent == 1.0:
            return Instruction(
                name="cz",
                qubits=tuple(qubits),
                args={},
            )
        raise OperationNotSupportedError(
            f"CZPowGate exponent was {operation.gate.exponent}, but only 1 is natively supported."
        )

    if isinstance(operation.gate, IQMMoveGate):
        return Instruction(
            name="move",
            qubits=tuple(qubits),
            args={},
        )

    if isinstance(operation, ClassicallyControlledOperation):
        if len(operation._conditions) > 1:
            raise OperationNotSupportedError("Classically controlled gates can currently only have one condition.")
        if not isinstance(operation._conditions[0], KeyCondition):
            raise OperationNotSupportedError("Only KeyConditions are supported as classical controls.")
        if isinstance(operation._sub_operation.gate, (PhasedXPowGate, XPowGate, YPowGate)):
            return Instruction(
                name="cc_prx",
                qubits=tuple(qubits),
                args={
                    "angle_t": operation._sub_operation.gate.exponent / 2,
                    "phase_t": operation._sub_operation.gate.phase_exponent / 2,
                    "feedback_qubit": "",
                    "feedback_key": str(operation._conditions[0]),
                },
            )
        raise OperationNotSupportedError(
            f"Classical control on the {operation._sub_operation.gate} gate is not supported."
        )

        # skipping feedback_qubit and feedback_key information until total circuit serialization

    raise OperationNotSupportedError(f"{type(operation.gate)} not natively supported.")


def serialize_circuit(circuit: Circuit) -> iqm_client.Circuit:
    """Serializes a quantum circuit into the IQM data transfer format.

    Args:
        circuit: quantum circuit to serialize

    Returns:
        data transfer object representing the circuit

    """
    total_ops_list = [op for moment in circuit for op in moment]
    cc_prx_support = any(isinstance(op, ClassicallyControlledOperation) for op in total_ops_list)
    instructions = list(map(map_operation, total_ops_list))

    if cc_prx_support:
        mkey_to_measurement = {}
        for inst in instructions:
            if inst.name == "measure":
                if inst.args["key"] in mkey_to_measurement:
                    raise OperationNotSupportedError("Cannot use the same key for multiple measurements.")
                mkey_to_measurement[inst.args["key"]] = inst
            elif inst.name == "cc_prx":
                feedback_key = inst.args["feedback_key"]
                measurement = mkey_to_measurement.get(feedback_key)
                if measurement is None:
                    raise OperationNotSupportedError(
                        f"cc_prx has feedback_key {feedback_key}, but no measure operation with that key precedes it."
                    )
                if len(measurement.qubits) != 1:
                    raise OperationNotSupportedError("cc_prx must depend on the measurement result of a single qubit.")
                inst.args["feedback_qubit"] = measurement.qubits[0]
                measurement.args["feedback_key"] = feedback_key

    return iqm_client.Circuit(name="Serialized from Cirq", instructions=instructions)


def deserialize_circuit(circuit: iqm_client.Circuit) -> Circuit:
    """Deserializes a quantum circuit from the IQM data transfer format to a Cirq Circuit.

    Args:
        circuit: data transfer object representing the circuit

    Returns:
        quantum circuit

    """
    return Circuit(
        map(
            instruction_to_operation,
            circuit.instructions,
        )
    )



# Copyright 2020–2022 Cirq on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""IQM's Adonis quantum architecture."""

from __future__ import annotations

from .iqm_device import IQMDevice, IQMDeviceMetadata


class Adonis(IQMDevice):
    """IQM's five-qubit transmon device.

    The qubits are connected thus::

            QB1
             |
      QB2 - QB3 - QB4
             |
            QB5

    where the lines denote which qubit pairs can be subject to two-qubit gates.

    Adonis has native PhasedXPowGate, XPowGate, and YPowGate gates. The two-qubit gate CZ is
    native, as well. The qubits can be measured simultaneously or separately any number of times.
    """

    def __init__(self):
        qubit_count = 5
        connectivity = ({1, 3}, {2, 3}, {4, 3}, {5, 3})
        super().__init__(IQMDeviceMetadata.from_qubit_indices(qubit_count, connectivity))



# Copyright 2023 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This file is an example of using Qiskit on IQM to run a simple but non-trivial quantum circuit on an IQM quantum
computer. See the Qiskit on IQM user guide for instructions:
https://docs.meetiqm.com/iqm-client/user_guide_qiskit.html
"""

import argparse

from iqm.qiskit_iqm.iqm_provider import IQMProvider
from qiskit import QuantumCircuit, transpile


def bell_measure(server_url: str) -> dict[str, int]:
    """Run a circuit that prepares and measures a Bell state.

    Args:
        server_url: URL of the IQM server used for execution

    Returns:
        a mapping of bitstrings representing qubit measurement results to counts for each result

    """
    backend = IQMProvider(server_url).get_backend()
    if backend.num_qubits < 2:
        raise ValueError("We need two qubits for the Bell state.")
    circuit = QuantumCircuit(2)
    circuit.h(0)
    circuit.cx(0, 1)
    circuit.measure_all()

    new_circuit = transpile(circuit, backend)
    return backend.run(new_circuit, shots=1000).result().get_counts()


if __name__ == "__main__":
    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        "--cortex_server_url",
        help="URL of the IQM server",
        default="https://demo.qc.iqm.fi/cocos",
    )
    print(bell_measure(argparser.parse_args().cortex_server_url))



# Copyright 2022-2025 Qiskit on IQM developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Error profile and fake backend base class for simulating IQM quantum computers."""

from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from itertools import permutations
from typing import Optional, Union
from uuid import UUID

from iqm.iqm_client import (
    DynamicQuantumArchitecture,
    GateImplementationInfo,
    GateInfo,
    Locus,
    StaticQuantumArchitecture,
)
from iqm.qiskit_iqm.iqm_backend import IQM_TO_QISKIT_GATE_NAME, IQMBackendBase
from iqm.qiskit_iqm.iqm_circuit_validation import validate_circuit
from iqm.qiskit_iqm.iqm_transpilation import IQMReplaceGateWithUnitaryPass
from iqm.qiskit_iqm.move_gate import MOVE_GATE_UNITARY
from qiskit import QuantumCircuit
from qiskit.providers import JobV1, Options
from qiskit_aer import AerSimulator
from qiskit_aer.noise import NoiseModel, QuantumError
from qiskit_aer.noise.errors import depolarizing_error, thermal_relaxation_error


def _dqa_from_sqa(
    sqa: StaticQuantumArchitecture,
    error_profile: IQMErrorProfile,
) -> DynamicQuantumArchitecture:
    """Create a dynamic quantum architecture from the given static quantum architecture
    and error profile.

    Since the DQA contains some attributes that are not present in an SQA or error profile,
    they are filled with mock data:

    * Each gate type is given a single mock implementation.
    * Calibration set ID is set to the all-zeros UUID.

    Args:
        sqa: Static quantum architecture to replicate.
        error_profile: Characteristics of a particular QPU specimen (including gate loci).

    Returns:
        DQA replicating the properties of ``sqa``.

    """

    def fake_gateinfo(loci: tuple[Locus, ...]) -> GateInfo:
        """Fake GateInfo for the given loci."""
        return GateInfo(
            implementations={"__fake": GateImplementationInfo(loci=loci)},
            default_implementation="__fake",
            override_default_implementation={},
        )

    # add all gates/loci found in the error profile
    gates = {
        gate_name: fake_gateinfo(tuple((component,) for component in gate_errors_1q))
        for gate_name, gate_errors_1q in error_profile.single_qubit_gate_depolarizing_error_parameters.items()
    }

    for gate_name, gate_errors_2q in error_profile.two_qubit_gate_depolarizing_error_parameters.items():
        gates[gate_name] = fake_gateinfo(tuple(gate_errors_2q))

    gates["measure"] = fake_gateinfo(tuple((component,) for component in error_profile.readout_errors))

    return DynamicQuantumArchitecture(
        calibration_set_id=UUID("00000000-0000-0000-0000-000000000000"),
        qubits=sqa.qubits,
        computational_resonators=sqa.computational_resonators,
        gates=gates,
    )


@dataclass
class IQMErrorProfile:
    r"""Characteristics of an IQM QPU specimen, used for constructing an error model.

    All the attributes of this class refer to the components of the QPU using their physical names.
    There are two types of QPU components, qubits and computational resonators.

    Args:
        t1s: maps components to their :math:`T_1` times (in ns)
        t2s: maps components to their :math:`T_2` times (in ns)
        single_qubit_gate_depolarizing_error_parameters: Depolarizing error parameters for single-qubit gates.
            Maps single-qubit gate names to a mapping of qubits (on which the gate acts) to a depolarizing error.
            The error, used in a one-qubit depolarizing channel, concatenated with a thermal relaxation channel,
            leads to average gate fidelities that would be determined by benchmarking.
        two_qubit_gate_depolarizing_error_parameters: Depolarizing error parameters for two-qubit gates.
            Maps two-qubit gate names to a mapping of pairs of qubits (on which the gate acts) to a depolarizing error.
            The error, used in a two-qubit depolarizing channel, concatenated with thermal relaxation channels for the
            qubits, leads to average gate fidelities that would be determined by benchmarking.
        single_qubit_gate_durations: Gate duration (in ns) for each single-qubit gate
        two_qubit_gate_durations: Gate duration (in ns) for each two-qubit gate.
        readout_errors: Maps physical qubit names to dicts that describe their single-qubit readout errors.
            For each qubit, the inner dict maps the state labels "0" and "1" to the probability :math:`P(\neg x|x)`
            of observing the state :math:`\ket{\neg x}` given the true state is :math:`\ket{x}`.
        name: Identifier of the QPU specimen.

    Example:
        .. code-block::

            IQMErrorProfile(
                t1s={"QB1": 10000.0, "QB2": 12000.0, "QB3": 14000.0},
                t2s={"QB1": 10000.0, "QB2": 12000.0, "QB3": 13000.0},
                single_qubit_gate_depolarizing_error_parameters={"prx": {"QB1": 0.0005, "QB2": 0.0004, "QB3": 0.0010}},
                two_qubit_gate_depolarizing_error_parameters={"cz": {("QB1", "QB2"): 0.08, ("QB2", "QB3"): 0.03}},
                single_qubit_gate_durations={"prx": 50.},
                two_qubit_gate_durations={"cz": 100.},
                readout_errors={"QB1": {"0": 0.02, "1": 0.03},
                                "QB2": {"0": 0.02, "1": 0.03},
                                "QB3": {"0": 0.02, "1": 0.03}},
                name="threequbit-example_sample"
            )

    """

    t1s: dict[str, float]
    t2s: dict[str, float]
    single_qubit_gate_depolarizing_error_parameters: dict[str, dict[str, float]]
    two_qubit_gate_depolarizing_error_parameters: dict[str, dict[tuple[str, str], float]]
    single_qubit_gate_durations: dict[str, float]
    two_qubit_gate_durations: dict[str, float]
    readout_errors: dict[str, dict[str, float]]
    name: Optional[str] = None

    def thermal_relaxation(self, component: str, duration: float) -> QuantumError:
        """One-qubit relaxation error channel."""
        return thermal_relaxation_error(self.t1s[component], self.t2s[component], duration)


class IQMFakeBackend(IQMBackendBase):
    """Simulated backend that mimics the behaviour of IQM quantum computers.

    Can be used to perform noisy gate-level simulations of quantum circuit execution on IQM hardware.

    A fake backend contains information about a specific IQM system, such as the static quantum architecture
    (number of qubits, connectivity), and a noise model based on system parameters such as relaxation (:math:`T_1`)
    and dephasing (:math:`T_2`) times, gate infidelities, and readout errors.

    Args:
        architecture: Static quantum architecture associated with the backend instance.
        error_profile: Characteristics of a particular QPU specimen.

    """

    def __init__(
        self,
        architecture: StaticQuantumArchitecture,
        error_profile: IQMErrorProfile,
        name: str = "IQMFakeBackend",
        **kwargs,
    ):
        self._validate_architecture_and_error_profile(architecture, error_profile)
        dqa = _dqa_from_sqa(architecture, error_profile)
        super().__init__(dqa, **kwargs)
        self.__sqa = architecture
        self.__dqa = dqa
        self.__error_profile = error_profile

        self.noise_model = self._create_noise_model()
        self.name = name

    @property
    def error_profile(self) -> IQMErrorProfile:
        """Error profile of this IQM fake backend instance."""
        return deepcopy(self.__error_profile)

    @error_profile.setter
    def error_profile(self, value: IQMErrorProfile) -> None:
        raise NotImplementedError(
            "Setting error profile of existing fake backend is not allowed. "
            "You may consider using the method .copy_with_error_profile."
        )

    def copy_with_error_profile(self, new_error_profile: IQMErrorProfile) -> IQMFakeBackend:
        """Return another instance of IQMFakeBackend, which has the same quantum architecture but a different error
        profile.
        """
        return self.__class__(self.__sqa, new_error_profile, self.name)

    @staticmethod
    def _validate_architecture_and_error_profile(
        architecture: StaticQuantumArchitecture, error_profile: IQMErrorProfile
    ) -> None:
        """Verify that the parameters of the QPU error profile match the constraints of its quantum architecture.

        Raises:
            ValueError: when length of :attr:`t1s` and number of qubits do not match.
            ValueError: when length of :attr:`t2s` and number of qubits do not match.
            ValueError: when length of :attr:`one_qubit_gate` parameter lists and number of qubits do not match.
            ValueError: when length of :attr:`two_qubit_gate` parameter lists and number of couplings do not match.
            ValueError: when gates in gate parameter lists are not supported by the quantum architecture.

        """

        def compare_sets(a: set, b: set, msg: str) -> None:
            """Compare the contents of two sets."""
            if diff := a - b:
                raise ValueError(f"Error profile attribute {msg} has unknown loci: {diff}.")
            if diff := b - a:
                raise ValueError(f"Error profile attribute {msg} is missing loci: {diff}.")

        components = set(architecture.qubits + architecture.computational_resonators)
        # Check that T1 and T2 dicts have one item for each component
        compare_sets(set(error_profile.t1s.keys()), components, "t1s")
        compare_sets(set(error_profile.t2s.keys()), components, "t2s")

        # Check that one-qubit gate parameter qubits match those of the architecture
        for gate, gate_errors_1q in error_profile.single_qubit_gate_depolarizing_error_parameters.items():
            compare_sets(
                set(gate_errors_1q.keys()),
                set(architecture.qubits),
                f"single_qubit_gate_depolarizing_error_parameters['{gate}']",
            )

        # Check that two-qubit gate parameter couplings match those of the architecture
        # Locus order is ignored because the loci in architecture.connectivity are unordered.
        for gate, gate_errors_2q in error_profile.two_qubit_gate_depolarizing_error_parameters.items():
            compare_sets(
                set(frozenset(locus) for locus in gate_errors_2q.keys()),
                set(frozenset(locus) for locus in architecture.connectivity),
                f"two_qubit_gate_depolarizing_error_parameters['{gate}']",
            )

        # Check that the basis gates in the error profile are all known
        known_gates = {"prx", "cc_prx", "cz", "move"}
        for property_name, specified_gates in [
            (
                "single_qubit_gate_depolarizing_error_parameters",
                error_profile.single_qubit_gate_depolarizing_error_parameters.keys(),
            ),
            (
                "two_qubit_gate_depolarizing_error_parameters",
                error_profile.two_qubit_gate_depolarizing_error_parameters.keys(),
            ),
            ("durations", (error_profile.single_qubit_gate_durations | error_profile.two_qubit_gate_durations).keys()),
        ]:
            for gate in specified_gates:
                if gate not in known_gates:
                    raise ValueError(f"Gate '{gate}' in {property_name} is unknown. Valid gates: {known_gates}")

        compare_sets(set(error_profile.readout_errors.keys()), set(architecture.qubits), "readout_errors")

    def _create_noise_model(self) -> NoiseModel:
        """Build a noise model from the attributes."""
        error_profile = self.__error_profile
        iqm_to_qiskit_gates = IQM_TO_QISKIT_GATE_NAME.copy()
        for iqm_gate in self.__dqa.gates:
            if iqm_gate not in ["measure", "barrier"]:
                iqm_to_qiskit_gates.setdefault(iqm_gate, iqm_gate)

        noise_model = NoiseModel(basis_gates=list(iqm_to_qiskit_gates.values()))

        # Add single-qubit gate errors to noise model
        for gate, gate_errors_1q in error_profile.single_qubit_gate_depolarizing_error_parameters.items():
            gate_duration = error_profile.single_qubit_gate_durations[gate]
            for component, gate_error in gate_errors_1q.items():
                thermal_relaxation_channel = error_profile.thermal_relaxation(component, gate_duration)
                depolarizing_channel = depolarizing_error(gate_error, 1)
                noise_model.add_quantum_error(
                    thermal_relaxation_channel.compose(depolarizing_channel),
                    iqm_to_qiskit_gates[gate],
                    [self.qubit_name_to_index(component)],
                )

        # Add two-qubit gate errors to noise model
        for gate, gate_errors_2q in error_profile.two_qubit_gate_depolarizing_error_parameters.items():
            gate_duration = error_profile.two_qubit_gate_durations[gate]
            for locus, gate_error in gate_errors_2q.items():
                for qb_order in permutations(locus):
                    # TODO why do we need to add the other locus order?
                    thermal_channels = [
                        error_profile.thermal_relaxation(component, gate_duration) for component in qb_order
                    ]
                    thermal_relaxation_channel = thermal_channels[0].tensor(thermal_channels[1])
                    depolarizing_channel = depolarizing_error(gate_error, 2)
                    noise_model.add_quantum_error(
                        thermal_relaxation_channel.compose(depolarizing_channel),
                        iqm_to_qiskit_gates[gate],
                        [self.qubit_name_to_index(qb_order[0]), self.qubit_name_to_index(qb_order[1])],
                    )

        # Add readout errors
        for qb, readout_error in error_profile.readout_errors.items():
            probabilities = [[1 - readout_error["0"], readout_error["0"]], [readout_error["1"], 1 - readout_error["1"]]]
            noise_model.add_readout_error(probabilities, [self.qubit_name_to_index(qb)])

        return noise_model

    @classmethod
    def _default_options(cls) -> Options:
        return Options(shots=1024)

    @property
    def max_circuits(self) -> Optional[int]:
        return None

    def run(self, run_input: Union[QuantumCircuit, list[QuantumCircuit]], **options) -> JobV1:
        """Run quantum circuits on the fake backend (by simulating them).

        This method will run the simulation with the noise model of the fake backend.
        Validity of the circuits is also checked.

        Args:
            run_input: One or more quantum circuits to simulate on the backend.
            options: Any kwarg options to pass to the backend.

        Returns:
            The job object representing the run.

        Raises:
            ValueError: Empty list of circuits was provided.

        """
        circuits_aux = [run_input] if isinstance(run_input, QuantumCircuit) else run_input

        if len(circuits_aux) == 0:
            raise ValueError("Empty list of circuits submitted for execution.")

        circuits = []
        GATE_TO_UNITARY = {
            "move": MOVE_GATE_UNITARY,
        }

        for circ in circuits_aux:
            validate_circuit(circ, self)
            circ_updated = circ

            for g in self.noise_model.basis_gates:
                if g not in IQM_TO_QISKIT_GATE_NAME.values():
                    circ_updated = IQMReplaceGateWithUnitaryPass(g, GATE_TO_UNITARY[g])(circ)
            circuits.append(circ_updated)

        shots = options.get("shots", self.options.shots)

        # Create noisy simulator backend and run circuits
        sim_noise = AerSimulator(noise_model=self.noise_model)

        job = sim_noise.run(circuits, shots=shots)
        return job

    def validate_compatible_architecture(self, architecture: StaticQuantumArchitecture) -> bool:
        """Compare a static quantum architecture to the static architecture of the fake backend.

        Args:
            architecture: static quantum architecture to compare to

        Returns:
            True iff the locus components and the component connectivity
            in the IQMFakeBackend SQA match ``architecture``.

        """
        return architecture == self.__sqa
